#!/usr/bin/env python

"""Tests for twotime correlation."""


import unittest
from boost_corr.xpcs_aps_8idi.xpcs_result import get_metadata, is_metadata
import os
import shutil
import tempfile
import h5py


class TestBoost_corr(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir = tempfile.TemporaryDirectory()
        temp_dir_fname = self.temp_dir.name
        flist = [os.path.join(temp_dir_fname, x + '.hdf') for x in [
            'false', 'true']]

        with h5py.File(flist[0], 'w') as f:
            f['/exchange/data'] = 1.0
        with h5py.File(flist[1], 'w') as f:
            f['/hdf_metadata_version'] = 2.0
        self.flist = flist

    def test_000_is_meta(self):
        self.assertEqual(is_metadata(self.flist[0]), False)
        self.assertEqual(is_metadata(self.flist[1]), True)
        self.assertEqual(is_metadata(''), False)
    
    def test_001_get_meta(self):
        meta_fname = get_metadata(self.temp_dir.name)
        self.assertEqual(meta_fname, self.flist[1])
    
    def test_002_not_found(self):
        os.remove(self.flist[1])
        self.assertRaises(FileNotFoundError,
                          lambda: get_metadata(self.temp_dir.name))

    def tearDown(self) -> None:
        self.temp_dir.cleanup()


if __name__ == '__main__':
    unittest.main()



key = {
    "delay_list": "/xpcs/multitau/delay_list",
    "g2": "/xpcs/multitau/normalized_g2",
    "g2_err": "/xpcs/multitau/normalized_g2_err",
    "g2_partial": "/xpcs/multitau/normalized_g2_partial",
    "g2_partial_err": "/xpcs/multitau/normalized_g2_partial_err",
    "g2_partial_labels": "/xpcs/multitau/normalized_g2_partial_labels",
    "G2IPIF": "/xpcs/multitau/unnormalized_G2",
    "multitau_config": "/xpcs/multitau/config",
    "saxs_2d": "/xpcs/temporal_mean/scattering_2d",
    "saxs_sq": "/xpcs/temporal_mean/scattering_sq",
    "saxs_1d": "/xpcs/temporal_mean/scattering_1d",
    "saxs_1d_segments": "/xpcs/temporal_mean/scattering_1d_segments",
    "intensity_vs_time": "/xpcs/spatial_mean/intensity_vs_time",
    "c2_prefix": "/xpcs/twotime",
    "c2_g2": "/xpcs/twotime/normalized_g2",
    "c2_g2_segments": "/xpcs/twotime/normalized_g2_segments",
    "c2_delay": "/xpcs/twotime/delay_list",
    "twotime_config": "/xpcs/twotime/config",
    "processed_bins": "/xpcs/twotime/processed_bins",
    "rawdata_path": "/xpcs/rawdata_path",
}



import h5py
import numpy as np
import matplotlib.pyplot as plt
import os

from imm_handler import ImmDataset


target_dir = "/scratch/cluster_results"

with h5py.File(target_dir + '/B981_multitau_eigen.hdf', 'r') as f:
    a = f['/exchange/pixelSum'][()]
    a_mask = f['/xpcs/mask'][()].astype(np.int8)
    a = a * a_mask
    sqmap = f["/xpcs/sqmap"][()]
    sqlist = f["/xpcs/sqlist"][()]
    saxs1d_eigen = f["/exchange/partition-mean-total"][()][0]


with h5py.File(target_dir + '/B981_2_10k_star_dynamic_0p1Hz_Strain1.01mm_Ampl0.005mm_att5_Lq0_001_0001-0800.hdf', 'r') as f:
    b = f['/exchange/pixelSum'][()]
    b_mask = f['/xpcs/mask'][()].astype(np.int8)
    b = b * b_mask
    saxs1d_gpu = f["/exchange/partition-mean-total"][()][0]

# read raw data directly
imm = ImmDataset('/scratch/xpcs_data_raw/B981_2_10k_star_dynamic_0p1Hz_Strain1.01mm_Ampl0.005mm_att5_Lq0_001/B981_2_10k_star_dynamic_0p1Hz_Strain1.01mm_Ampl0.005mm_att5_Lq0_001_00001-00800.imm')
tot = np.zeros(imm.det_size, dtype=np.int64)
for n in range(imm.batch_num):
    x = imm[n].cpu().numpy().astype(np.int64).sum(axis=0)
    tot += x.reshape(imm.det_size)
tot = tot * a_mask
tot = tot / 1.0 / imm.frame_num

non_zero = np.nonzero(a_mask)
sl_v = slice(np.min(non_zero[0]), np.max(non_zero[0]) + 1)
sl_h = slice(np.min(non_zero[1]), np.max(non_zero[1]) + 1)

a = a[sl_v, sl_h]
b = b[sl_v, sl_h]
mask = a_mask[sl_v, sl_h]
tot = tot[sl_v, sl_h]

print(np.max(a), np.max(b), np.max(tot))
print(np.mean(a), np.mean(b), np.mean(tot))
sqmap = sqmap[sl_v, sl_h]

sqmap1d = sqmap.flatten()
tot1d = tot.flatten()
saxs1d_raw = np.bincount(sqmap1d, weights=tot1d)[1:]

count = np.bincount(sqmap1d)[1:]
valid_idx = count > 0
print(valid_idx.dtype, np.sum(valid_idx))
print("invalid_idx", np.nonzero(~valid_idx))
saxs1d_raw = saxs1d_raw[valid_idx] / count[valid_idx]

# print(saxs1d_eigen.shape)
# print(saxs1d_eigen[-20:])
# print(saxs1d_gpu[-20:])

xline = np.arange(saxs1d_eigen.size)
plt.semilogy(xline, saxs1d_eigen, 'rx-', ms=5, label='eigen', alpha=0.9)
plt.semilogy(xline, saxs1d_gpu, 'b+-', ms=5, label='gpu', alpha=0.9)
plt.semilogy(xline, saxs1d_raw, 'go-', alpha=0.5, ms=5, label='raw')
plt.legend()
plt.show()
plt.savefig('saxs1d_all.png', dpi=600)
plt.close()

# plt.plot(xline, saxs1d_eigen / saxs1d_raw - 1, 'rx-', label='eigen/raw - 1')
# plt.legend()
# plt.savefig('saxs1d_eigen.png', dpi=600)
# plt.close()
# 
# plt.plot(xline, saxs1d_gpu / saxs1d_raw - 1, 'bo-', label='gpu/raw - 1')
# plt.legend()
# plt.savefig('saxs1d_gpu.png', dpi=600)
# plt.close()
# saxs 2d comparison
# def find_vmin_vmax(diff):
#     vmin = np.min(diff)
#     vmax = np.max(diff)
#     max_diff = max(abs(vmin), vmax)
#     return -max_diff, max_diff
# 
# 
# diff = [a - b, tot - b, tot - a]
# # diff = [a / b - 1, np.divide tot / b - 1, tot /a - 1]
# label = ['eigen-gpu', 'raw-gpu', 'raw-eigen']
# 
# for n in range(3):
#     x = diff[n]
#     vmin, vmax = find_vmin_vmax(x)
#     plt.imshow(x, cmap=plt.get_cmap('seismic'), vmin=vmin, vmax=vmax)
#     plt.colorbar()
#     plt.title(label[n])
#     plt.savefig(label[n] + '.png', dpi=600)
#     plt.close()
# 


import os
import struct
import numpy as np


def convert_sparse(a):
    """
    convert sparse data from Rigaku 64bit binary format to 3xN format
    Args:
        a: 1D array of uint64
    Returns:
        3xN array of uint32, where N is the number of pixels
    """
    output = np.zeros(shape=(3, a.size), dtype=np.uint32)
    # index of pixels in the detector
    output[0] = ((a >> 16) & (2 ** 21 - 1)).astype(np.uint32)
    # index of frame
    output[1] = (a >> 40).astype(np.uint32)
    # photon count
    output[2] = (a & (2 ** 12 - 1)).astype(np.uint8)
    return output


def get_number_of_frames_from_binfile(filepath, endianness='<'):
    """Reads the last 8 bytes of a binary file and converts them to a NumPy uint64.
    Args:
        filepath: The path to the binary file.
        endianness:  '>' for big-endian, '<' for little-endian. Defaults to big-endian.
    Returns:
        Number of frames for a Rigaku binary file
    """
    file_size = os.path.getsize(filepath)
    if file_size < 8:
        raise ValueError("File size is less than 8 bytes. Cannot read the last 8 bytes.")

    with open(filepath, "rb") as f:
        f.seek(file_size - 8)
        last_8_bytes = f.read(8)

    format_string = endianness + "Q"  # Q for unsigned long long (8 bytes)
    value = struct.unpack(format_string, last_8_bytes)[0]
    # Convert to NumPy uint64.  This is important for consistent type handling!
    uint64_value = np.array(np.uint64(value)).reshape(1)
    # number of frames is the second element of the output of convert_sparse
    # add 1 to get the number of frames out of zero-indexing
    number_frames = int(convert_sparse(uint64_value)[1] + 1)
    return number_frames



#!/usr/bin/env python

"""Tests for twotime correlation."""


import unittest
from boost_corr import TwotimeCorrelator
import torch
import numpy as np


def run_and_get_result(frame_num=512, device='cpu', smooth_method='sqmap'):
    det_size = (4, 5)
    qinfo = {}
    qinfo['dq_idx'] = np.arange(1, 5)    # 1, 2, 3, 4
    qinfo['dq_slc'] = [slice(x * 5 - 5, x * 5) for x in qinfo['dq_idx']]
    qinfo['sq_idx'] = np.arange(1, 21)   # 1, 2, ... ,20
    qinfo['sq_slc'] = [slice(x - 1, x) for x in qinfo['sq_idx']]

    num_elm = det_size[0] * det_size[1]
    tt = TwotimeCorrelator(qinfo=qinfo, det_size=det_size, 
                           frame_num=frame_num, device=device)
    for _ in range(frame_num):
        x = torch.ones(det_size, device=device).reshape(-1, num_elm)
        tt.process(x)
    tt.post_processing(smooth_method=smooth_method)
    result_saxs = tt.get_saxs()
    result_twotime = tt.get_twotime_result()
    return result_saxs, result_twotime


class TestBoost_corr(unittest.TestCase):
    """Tests for `boost_corr` package."""

    def test_000_cpu_sqmap(self):
        frame_num = 1024
        saxs, twotime = run_and_get_result(frame_num=frame_num, device='cpu',
                                           smooth_method='sqmap')

        saxs2d = saxs['saxs2d']
        saxs2d_real = torch.ones_like(saxs2d)
        self.assertEqual(torch.allclose(saxs2d, saxs2d_real), True)

        saxs2d_par = saxs['saxs2d_par']
        saxs2d_par_real = torch.ones_like(saxs2d)
        self.assertEqual(torch.allclose(saxs2d_par, saxs2d_par_real), True)

        for d in twotime:
            for k, v in d.items():
                if '/exchange/C2T_all' in k:
                    v_real = np.zeros_like(v)
                    v_real[np.triu_indices(v_real.shape[0])] = 1
                    self.assertEqual(np.allclose(v, v_real), True)

    def test_001_cpu_dqmap(self):
        frame_num = 1024
        saxs, twotime = run_and_get_result(frame_num=frame_num, device='cpu',
                                           smooth_method='dqmap')

        saxs2d = saxs['saxs2d']
        saxs2d_real = torch.ones_like(saxs2d)
        self.assertEqual(torch.allclose(saxs2d, saxs2d_real), True)

        saxs2d_par = saxs['saxs2d_par']
        saxs2d_par_real = torch.ones_like(saxs2d)
        self.assertEqual(torch.allclose(saxs2d_par, saxs2d_par_real), True)

        for d in twotime:
            for k, v in d.items():
                if '/exchange/C2T_all' in k:
                    v_real = np.zeros_like(v)
                    v_real[np.triu_indices(v_real.shape[0])] = 1
                    self.assertEqual(np.allclose(v, v_real), True)

    def test_002_gpu(self):
        frame_num = 1024
        saxs, twotime = run_and_get_result(frame_num=frame_num, 
                                           device='cuda:0',
                                           smooth_method='sqmap')

        saxs2d = saxs['saxs2d']
        saxs2d_real = torch.ones_like(saxs2d)
        self.assertEqual(torch.allclose(saxs2d, saxs2d_real), True)

        saxs2d_par = saxs['saxs2d_par']
        saxs2d_par_real = torch.ones_like(saxs2d)
        self.assertEqual(torch.allclose(saxs2d_par, saxs2d_par_real), True)

        for d in twotime:
            for k, v in d.items():
                if '/exchange/C2T_all' in k:
                    v_real = np.zeros_like(v)
                    v_real[np.triu_indices(v_real.shape[0])] = 1
                    self.assertEqual(np.allclose(v, v_real), True)


if __name__ == '__main__':
    unittest.main()



import glob
import logging
import os
import shutil
import traceback
import h5py

from .append_metadata_qmap import append_metadata_qmap
from .hdf_reader import put_results_in_hdf5

logger = logging.getLogger(__name__)


def is_metadata(fname: str):
    if not os.path.isfile(fname):
        return False, None

    with h5py.File(fname, "r") as f:
        if "/entry/schema_version" in f:
            return True, "nexus"

    return False, None


def get_metadata(meta_dir: str):
    meta_fnames = glob.glob(meta_dir + "/*_metadata.hdf")
    if len(meta_fnames) >= 1:
        for f in meta_fnames:
            is_meta, meta_type = is_metadata(f)
            if is_meta:
                return f, meta_type
    raise FileNotFoundError(f"no metadata file found in [{meta_dir}]")


def create_unique_file(
    output_dir: str,
    meta_fname: str,
    overwrite: bool = False,
    prefix: str = None,
    suffix: str = None,
) -> str:
    """
    Create a unique filename for an HDF file, avoiding overwriting existing files.

    This function generates a unique filename based on the input parameters. It ensures
    the file has an .hdf extension, incorporates the analysis type if specified, and
    avoids overwriting existing files by appending a number to the filename if necessary.

    Args:
        output_dir (str): The directory where the file will be created.
        meta_fname (str): The base filename to use.
        analysis_type (Optional[str], optional): The type of analysis, if 'Twotime' is
            specified, '_Twotime' will be added to the filename. Defaults to None.
        overwrite (bool, optional): If True, allows overwriting existing files.
            Defaults to False.

    Returns:
        str: A unique filename (including path) that can be used to create a new file
             without overwriting existing files, unless overwrite is True.

    Examples:
        >>> create_unique_file('/output', 'data.txt')
        '/output/data.hdf'
        >>> create_unique_file('/output', 'data.hdf', analysis_type='Twotime')
        '/output/data_Twotime.hdf'
        >>> create_unique_file('/output', 'data.hdf', overwrite=False)
        '/output/data_01.hdf'  # if 'data.hdf' already exists
    """
    # Create the base filename
    base_fname = os.path.basename(meta_fname)

    # Ensure the filename has the .hdf extension
    name, ext = os.path.splitext(base_fname)
    name = name.rstrip("_metadata")

    if prefix:
        name = f"{prefix.rstrip('_')}_{name}"
    if suffix:
        name = f"{name}_{suffix.lstrip('_')}"

    base_fname = name + "_results.hdf"

    # Create the full path
    fname = os.path.join(output_dir, base_fname)

    # If overwrite is True, return the filename as is
    if overwrite or not os.path.isfile(fname):
        return fname

    # If the file exists and overwrite is False, create a unique filename
    name, ext = os.path.splitext(fname)
    counter: int = 1
    while True:
        new_fname = f"{name}_{counter}{ext}"
        if not os.path.isfile(new_fname):
            return new_fname
        counter += 1


class XpcsResult:
    def __init__(
        self,
        meta_dir=None,
        qmap_fname=None,
        output_dir=None,
        overwrite=False,
        rawdata_path=None,
        multitau_config=None,
        twotime_config=None,
        prefix=None,
        suffix=None,
    ) -> None:
        self.meta_dir = meta_dir
        self.qmap_fname = qmap_fname
        self.output_dir = output_dir
        self.overwrite = overwrite
        self.fname = None
        self.G2_fname = None
        self.fname_temp = None
        self.G2_fname_temp = None
        self.success = True
        self.prefix = prefix
        self.suffix = suffix
        self.analysis_config = {
            "rawdata_path": rawdata_path,
            "multitau_config": multitau_config or {},
            "twotime_config": twotime_config or {},
        }

    def __enter__(self):
        """
        Perform setup when entering the context manager.
        """
        meta_fname, meta_ftype = get_metadata(self.meta_dir)
        logger.info(f"metadata filename/type is {meta_fname} | {meta_ftype}")
        self.fname = create_unique_file(
            self.output_dir,
            meta_fname,
            overwrite=self.overwrite,
            prefix=self.prefix,
            suffix=self.suffix,
        )
        self.G2_fname = os.path.splitext(self.fname)[0] + "_G2.hdf"
        self.fname_temp = self.fname + ".temp"
        self.G2_fname_temp = self.G2_fname + ".temp"

        if not os.path.isdir(self.output_dir):
            os.makedirs(self.output_dir)
        append_metadata_qmap(self.fname_temp, meta_fname, self.qmap_fname)
        self.append(self.analysis_config)
        return self

    def append(self, result_dict):
        """
        Save the provided result_dict to the temporary files.
        """
        try:
            if "G2" in result_dict:
                self.append_and_link_G2(result_dict["G2"])
                del result_dict["G2"]
            put_results_in_hdf5(self.fname_temp, result_dict)
        except Exception:
            traceback.print_exc()
            self.success = False
            raise
        else:
            self.success = True and self.success

    def correct_t0_for_timepix4(self, t0):
        """
        Correct the t0 value in the multitau configuration for Timepix4 data.
        """
        logger.info(f"Correcting t0 for Timepix4 data. {t0=} s")
        with h5py.File(self.fname_temp, "r+") as f:
            if "/entry/instrument/detector_1/frame_time" in f:
                del f["/entry/instrument/detector_1/frame_time"]
            f["/entry/instrument/detector_1/frame_time"] = t0

    def append_and_link_G2(self, G2_dict):
        """
        append the provided result_dict to the temporary files.
        """
        try:
            put_results_in_hdf5(self.G2_fname_temp, {"G2IPIF": G2_dict}, mode="raw")
            # Create a link for the main file
            with h5py.File(self.fname_temp, "r+") as f:
                relative_path = os.path.basename(self.G2_fname)
                f["/xpcs/multitau/unnormalized_g2"] = h5py.ExternalLink(
                    relative_path, "/G2IPIF"
                )
        except Exception:
            traceback.print_exc()
            self.success = False
            raise
        else:
            self.success = True and self.success

    def __exit__(self, exc_type, exc_value, traceback_obj):
        """
        Exit the context manager, finalizing the temporary files.
        """
        if self.success:
            try:
                if os.path.isfile(self.fname_temp):
                    shutil.move(self.fname_temp, self.fname)
                if os.path.isfile(self.G2_fname_temp):
                    shutil.move(self.G2_fname_temp, self.G2_fname)
            except Exception:
                traceback.print_exc()
                logger.error("failed to rename the result files")
                self.success = False
                raise

        # try to clean up regardless of success
        for temp_file in [self.fname_temp, self.G2_fname_temp]:
            if os.path.exists(temp_file):
                os.remove(temp_file)
                logger.info(f"Cleaned up temporary file: {temp_file}")

        if self.success:
            logger.info(f"Succeeded in saving result file: {self.fname}")
        else:
            logger.error(f"Failed in saving result file: {self.fname}")
        return self.success


if __name__ == "__main__":
    pass



import torch
from tqdm import trange, tqdm
from torch.utils.data import DataLoader
import logging


logger = logging.getLogger(__name__)


def create_diagonal_index(N=512, device='cpu'):
    """
    create a matrix of indices for diagonal averaging
    :param N: size of the matrix
    :return: NxN matrix with indices for diagonal averaging
    Example:
    >>> create_diagonal_index(5)
    tensor([[0, 1, 2, 3, 4],
            [1, 0, 1, 2, 3],
            [2, 1, 0, 1, 2],
            [3, 2, 1, 0, 1],
            [4, 3, 2, 1, 0]])
    """
    indices = torch.arange(N, device=device)
    diag_mat = torch.abs(indices.unsqueeze(0) - indices.unsqueeze(1))
    return diag_mat


def compute_diagonal_average(index, weights):
    """
    compute the diagonal average of a matrix using precomputed indices and weights
    :param index:  matrix with indices for diagonal averaging
    :param weights: NxN matrix with weights for diagonal averaging
    """
    # index and weights must be both one d;
    tot = torch.bincount(index, weights=weights)
    pts = torch.bincount(index)
    return tot / pts


class TwotimeCorrelator():
    def __init__(self,
                 qinfo,
                 frame_num,
                 det_size=(1024, 512),
                 window=1024,
                 mask_crop=None,
                 device='cpu',
                 method='normal',
                 dtype=torch.float32) -> None:

        self.dq_idx = qinfo['dq_idx']
        self.dq_slc = qinfo['dq_slc']
        self.sq_idx = qinfo['sq_idx']
        self.sq_slc = qinfo['sq_slc']
        # self.dq_sq_map = qinfo['dq_sq_map']

        self.det_size = det_size
        self.pixel_num = self.det_size[0] * self.det_size[1]

        self.device = device
        self.frame_num = frame_num
        self.method = method

        if mask_crop is None:
            mask_crop = torch.ones(det_size, device=self.device,
                                   dtype=torch.bool)
            mask_crop = mask_crop.reshape(-1)
        self.mask_crop = mask_crop

        arr_size = self.dq_slc[-1].stop
        self.arr_size = arr_size

        num_dq = len(self.dq_slc)
        if self.method == 'normal':
            shape = (frame_num, arr_size)
        elif self.method == 'window':
            shape = (window * 2, arr_size)
            self.c2 = torch.zeros((num_dq, frame_num - window + 1, window),
                                  device=device)
            self.c2_ptr = 0
            # self.cache_sum = torch.zeros((window * 2, num_q), device=device)

        self.cache = torch.zeros(shape, device=device, dtype=dtype)
        self.cache_ptr = 0

        self.window = window

        self.frame_sum = None
        self.pixel_sum = None
        self.g2full = []
        self.g2partial = []
        # smooth data
        self.sdata = None
        self.c2_idx = 1

    def process(self, x):
        if self.method == 'normal':
            sz = x.shape[0]
            self.cache[self.cache_ptr:self.cache_ptr + sz] = x
            self.cache_ptr += sz
            # if self.cache_ptr == self.num_frames:
            #     self.calc_normal_twotime()
        elif self.method == 'window':
            self.process_window(x)
    
    def process_dataset(self,
                        ds,
                        verbose=True,
                        use_loader=False,
                        num_workers=16):
        if not use_loader:
            xrange = trange if verbose else range
            for n in xrange(len(ds)):
                self.process(ds[n])
        else:
            # only pin memory if using GPU for computing
            pin_memory = (self.device != 'cpu')
            dl = DataLoader(ds,
                            batch_size=None,
                            pin_memory=pin_memory,
                            num_workers=num_workers,
                            prefetch_factor=4)
            logger.info(f'using {num_workers} workers to load data')
            
            if verbose:
                container = tqdm(dl, total=len(ds), desc='progress')
            else:
                container = dl

            for x in container:
                x = x.to(self.device, non_blocking=True)
                self.process(x)

    def compute_average(self, num_part=10):
        pixel_sum_par = torch.zeros(num_part,
                                    self.pixel_num,
                                    dtype=torch.float32,
                                    device=self.device)
        self.pixel_sum = torch.zeros_like(pixel_sum_par[0])

        frame_len = self.frame_num // num_part
        for n in range(num_part):
            if n == num_part - 1:
                sl = slice(n * frame_len, self.frame_num)  # last part
            else:
                sl = slice(n * frame_len, (n + 1) * frame_len)
            part_sum = torch.sum(self.cache[sl], axis=0)
            self.pixel_sum[self.mask_crop] += part_sum
            pixel_sum_par[n, self.mask_crop] = part_sum / (sl.stop - sl.start)

        self.pixel_sum /= self.frame_num
        self.pixel_sum = self.pixel_sum.reshape(*self.det_size)
        self.pixel_sum_par = pixel_sum_par.reshape(num_part, *self.det_size)

        frame_sum_val = torch.mean(self.cache, axis=1)
        # self.cache /= frame_sum_val.reshape(-1, 1)

        frame_sum_xax = torch.arange(self.frame_num, device=self.device)
        self.frame_sum = torch.stack([frame_sum_xax, frame_sum_val])

    def compute_smooth_data(self, mode='sqmap'):
        if mode == 'sqmap':
            slc = self.sq_slc
        elif mode == 'dqmap':
            slc = self.dq_slc
        else:
            raise ValueError('smooth method not supported')

        cache = self.cache.T
        for n in range(len(slc)):
            avg = cache[slc[n]].mean(dim=0)
            # get rid of the zeros
            avg[avg <= 0] = 1
            cache[slc[n]] /= avg

        self.cache = cache.T
        return

    def process_window(self, x):
        if self.cache_ptr >= 2 * self.window:
            self.cache_ptr = self.window
            self.cache = torch.roll(self.cache, shifts=-self.window, dims=0)

        x0 = x[0]
        self.cache[self.cache_ptr] = x0
        self.cache_ptr += 1

        if self.cache_ptr >= self.window:
            sl = slice(self.cache_ptr - self.window, self.cache_ptr)
            wf = self.cache[sl]
            for idx, roi in enumerate(self.dq_slc):
                a = wf[:, roi]
                b = x0[roi]
                a_sum = torch.sum(a, dim=1)
                b_sum = torch.sum(b)
                corr = torch.matmul(a, b.T) * len(b) / b_sum / a_sum
                self.c2[idx, self.c2_ptr] = corr
            self.c2_ptr += 1

        if x.shape[0] <= 1:
            return
        else:
            self.process_window(x[1:])
            return

    def post_processing(self, smooth_method=None, **kwargs):
        self.cache = self.cache.float()
        # saxs2d
        self.compute_average()
        # smooth data
        self.compute_smooth_data(smooth_method)

    def get_twotime_generator(self, **kwargs):
        for c2 in self.calc_normal_twotime(**kwargs):
            yield {f'correlation_map/c2_{self.c2_idx:05d}': c2}
            self.c2_idx += 1

        self.g2full = torch.stack(self.g2full).swapaxes(0, 1)
        self.g2partial = torch.stack(self.g2partial).permute(2, 1, 0)
        results = {
            'c2_g2': self.g2full,
            'c2_g2_segments': self.g2partial,
            'processed_bins': self.dq_idx
        }

        for k, v in results.items():
            if isinstance(v, torch.Tensor):
                results[k] = v.cpu().numpy()
        yield results

    def get_scattering(self):
        num_par = self.pixel_sum_par.shape[0]
        saxs = {
            'saxs_2d': self.pixel_sum.reshape(-1)[self.mask_crop],
            'saxs_2d_par': self.pixel_sum_par.reshape(num_par, -1)[:, self.mask_crop],
            'mask_crop': self.mask_crop,
            'intensity_vs_time': self.frame_sum,
        }
        return saxs

    def calc_normal_twotime(self, num_partials=5):
        diag_mat = create_diagonal_index(self.frame_num, device=self.device)
        diag_mat_1d = diag_mat.reshape(-1)
        partial_len = self.frame_num // num_partials
        diag_mat_1d_par = diag_mat[0:partial_len, 0:partial_len].reshape(-1)

        # triu_idx = torch.tril_indices(self.num_frames, self.num_frames)
        for n, _ in enumerate(self.dq_idx):
            time_series = self.cache[:, self.dq_slc[n]]
            if self.sdata is not None:
                time_series = time_series / self.sdata[self.dq_slc[n]]

            norm_factor = torch.sum(time_series, dim=1)
            norm_factor[norm_factor <= 0] = 1
            norm_factor = 1.0 / norm_factor

            matmul_prod = torch.matmul(time_series, time_series.T)
            c2 = matmul_prod * norm_factor * norm_factor.reshape(-1, 1) * time_series.shape[1]

            g2full = compute_diagonal_average(diag_mat_1d, c2.reshape(-1))
            self.g2full.append(g2full)

            g2partial = []
            for idx in range(num_partials):
                sl = slice(idx * partial_len, (idx + 1) * partial_len)
                temp = compute_diagonal_average(diag_mat_1d_par,
                                                c2[sl, sl].reshape(-1))
                g2partial.append(temp)
            g2partial = torch.stack(g2partial)
            self.g2partial.append(g2partial)

            yield (torch.triu(c2)).cpu().numpy()


if __name__ == '__main__':
    pass



"""Top-level package for boost-corr."""

__author__ = """Miaoqi Chu"""
__email__ = "mqichu@anl.gov"

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "0.0.0"

import logging

logging.getLogger("boost_corr").addHandler(logging.NullHandler())

from .multitau import MultitauCorrelator
from .twotime import TwotimeCorrelator

__all__ = (MultitauCorrelator, TwotimeCorrelator)



from boost_corr.xpcs_aps_8idi.rigaku_3M_handler import Rigaku3MDataset

fname = '/home/8ididata/Rigaku3M_Test_Data/SparsificationData2/Div1_000001.bin.000'
dset = Rigaku3MDataset(fname, batch_size=1000, device='cuda:0')
print(dset.det_size)
print(dset.frame_num)
# plt.imshow(dset.get_scattering())
# plt.show()

for n in range(100):
    x = dset.__getbatch__(n)
    print(n, x.shape, x.device)




import numpy as np

# required for bloc compression
import hdf5plugin
import h5py
import torch
import logging
from .xpcs_dataset import XpcsDataset


logger = logging.getLogger(__name__)


class HdfDataset(XpcsDataset):
    """
    Parameters
    ----------
    filename: string
    """
    def __init__(self,
                 *args,
                 preload_size=8,
                 dtype=np.uint8,
                 data_path='/entry/data/data',
                 **kwargs):

        super(HdfDataset, self).__init__(*args, dtype=dtype, **kwargs)
        self.dataset_type = "HDF5 Dataset"
        self.is_sparse = False
        with h5py.File(self.fname, 'r') as f:
            data = f[data_path]
            self.shape = data.shape

            # update data type;
            if data.dtype == np.uint8:
                self.dtype = np.int16
            elif data.dtype == np.uint16:
                # lambda2m's uint16 is actually 12bit. it's safe to to int16
                if self.shape[1] * self.shape[2] == 1813 * 1558:
                    self.dtype = np.int16
                # likely eiger detectors
                else:
                    self.dtype = np.int32
            elif data.dtype == np.uint32:
                self.dtype = np.int32
                logger.warn('cast uint32 to int32. it may cause ' + 
                            'overflow when the maximal value >= 2^31')
            else:
                self.dtype = data.dtype
        
        if kwargs['avg_frame'] > 1:
            self.dtype = np.float32
        self.fhdl = None
        self.data = None
        self.data_cache = None
        self.data_path = data_path

        self.update_batch_info(self.shape[0])
        self.update_det_size(self.shape[1:])

        if self.mask_crop is not None:
            self.mask_crop = self.mask_crop.cpu().numpy()

        self.current_group = None
        self.preload_size = preload_size
        self.cache = None

    def __reset__(self):
        if self.fhdl is not None:
            self.fhdl.close()
            self.data = None
            self.fhdl = None

    def __getbatch__(self, idx):
        """
        return numpy array, which will be converted to tensor with dataloader
        """
        if self.fhdl is None:
            self.fhdl = h5py.File(self.fname, 'r', rdcc_nbytes=1024*1024*256)
            self.data = self.fhdl[self.data_path]

        beg, end, size = self.get_raw_index(idx)
        if self.cache is None or self.cache.shape[0] != size:
            self.cache = np.zeros(shape=(size, self.shape[1], self.shape[2]), dtype=self.dtype)
        # idx_list = np.arange(beg, end, self.stride)

        self.data.read_direct(self.cache, np.s_[beg:end:self.stride])
        x = self.cache.reshape(size, self.pixel_num)
        if self.mask_crop is not None:
            x = x[:, self.mask_crop]

        return torch.from_numpy(x)


def test():
    fname = (
        "/clhome/MQICHU/ssd/xpcs_data_raw/A003_Cu3Au_att0_001/A003_Cu3Au_att0_001.imm"
    )
    ds = HdfDataset(fname)
    # for n in range(len(ds)):
    #     print(n, ds[n].shape, ds[n].device)


def test_bin(idx):
    fname = f"/clhome/MQICHU/ssd/APSU_TestData_202106/APSU_TestData_{idx:03d}/APSU_TestData_{idx:03d}.h5"
    ds = HdfDataset(fname)
    ds.to_rigaku_bin(f"hdf2bin_{idx:03d}.bin")


if __name__ == '__main__':
    test_bin(4)



import numpy as np

nexus_xpcs_schema = {
    "twotime": {
        "type": "NXdata",
        "required": False,
        "description": "The data (results) in this section are based on the two-time intensity correlation function derived from a time series of scattering images",
        "g2_err_from_two_time_corr_func": {
            "type": "NX_NUMBER",
            "units": "NX_DIMENSIONLESS",
            "required": False,
            "description": "error values for the g2 values",
            "data": np.zeros((2, 3)),
            "attributes": {
                "storage_mode": {
                    "type": "NX_CHAR", 
                    "required": True,
                    "data": "one_array_q_first",
                    "description": "Any of these values: one_array_q_first one_array_q_last data_exchange_keys other"
                }
            }
        },
        "g2_from_two_time_corr_func": {
            "type": "NX_NUMBER",
            "units": "NX_DIMENSIONLESS",
            "required": False,
            "data": np.zeros((2, 3)),
            "description": "frame weighted average along the diagonal direction in two_time_corr_func",
            "attributes": {
                "storage_mode": {
                    "type": "NX_CHAR", 
                    "required": True,
                    "data": "one_array_q_first",
                    "description": "Any of these values: one_array_q_first one_array_q_last data_exchange_keys other"
                },
                "baseline_reference": {
                    "type": "NX_INT", 
                    "required": True,
                    "data": 1.0,
                    "description": "Any of these values: 0 | 1"
                },
                "first_point_for_fit": {
                    "type": "NX_INT", 
                    "required": True,
                    "data": 0,
                    "description": "first_point_for_fit describes if the first point should or should not be included in fitting"
                }
            }
        },
        "two_time_corr_func": {
            "type": "NX_NUMBER",
            "units": "NX_ANY",
            "required": False,
            "description": "two-time correlation of speckle intensity for a given q-bin or roi",
            "data": np.zeros((2, 3)),
            "attributes": {
                "storage_mode": {
                    "type": "NX_CHAR", 
                    "required": True,
                    "data": "one_array_q_first",
                    "description": "storage_mode describes the format of the data to be loaded"
                },
                "baseline_reference": {
                    "type": "NX_INT", 
                    "required": True,
                    "data": 1,
                    "description": "baseline is the expected value of a full decorrelation"
                },
                "populated_elements": {
                    "type": "NX_CHAR", 
                    "required": True,
                    "data": "upper_half",
                    "description": "populated_elements describe the elements of the 2D array that are populated"
                },
                "time_origin_location": {
                    "type": "NX_CHAR", 
                    "required": True,
                    "data": "upper_left",
                    "description": "time_origin_location is the location of the origin"
                }
            }
        }
    },
    "multitau": {
        "type": "NXdata",
        "required": True,
        "description": "The multitau results data captured here are most commonly required for high throughput, equilibrium dynamics experiments",
        "delay_difference": {
            "type": "NX_INT",
            "units": "NX_COUNT",
            "required": True,
            "data": 1,
            "description": "delay_difference (also known as delay or lag step)",
            "attributes": {
                "storage_mode": {
                    "type": "NX_CHAR", 
                    "required": True,
                    "data": "one_array_q_first",
                    "description": "Any of these values: one_array | data_exchange_keys | other"
                }
            }
        },
        "frame_average": {
            "type": "NX_NUMBER", 
            "units": "NX_COUNT", 
            "required": True,
            "description": "Two-dimensional average along the frames stack",
            "data": 1
        },
        "frame_sum": {
            "type": "NX_NUMBER", 
            "units": "NX_COUNT", 
            "required": True,
            "description": "Two-dimensional summation along the frames stack",
            "data": 1,
        },
        "g2": {
            "type": "NX_NUMBER",
            "units": "NX_DIMENSIONLESS",
            "required": True,
            "description": "normalized intensity auto-correlation function",
            "data": None,
            "attributes": {
                "storage_mode": {
                    "type": "NX_CHAR", 
                    "required": True,
                    "data": "one_array_q_first",
                    "description": "storage_mode describes the format of the data to be loaded"
                }
            }
        },
        "g2_derr": {
            "type": "NX_NUMBER",
            "units": "NX_DIMENSIONLESS",
            "required": True,
            "description": "error values for the g₂ values",
            "data": None,
            "attributes": {
                "storage_mode": {
                    "type": "NX_CHAR", 
                    "required": True,
                    "data": "one_array_q_first",
                    "description": "Any of these values: one_array | data_exchange_keys | other"
                }
            }
        },
        "G2_unnormalized": {
            "type": "NX_NUMBER",
            "units": "NX_ANY",
            "required": True,
            "description": "unnormalized intensity auto-correlation function",
            "data": None,
            "attributes": {
                "storage_mode": {
                    "type": "NX_CHAR", 
                    "required": True,
                    "data": "one_array_q_first",
                    "description": "Any of these values: one_array | data_exchange_keys | other"
                }
            }
        }
    },
    
}


import os
import sys
import time
import random
import signal
from datetime import datetime
from uuid import uuid4
from filelock import FileLock, Timeout
import pynvml
import logging


logger = logging.getLogger(__name__)


class GPUScheduler:
    LOCK_DIR = "/dev/shm/gpu_locks"
    QUEUE_DIR = "/dev/shm/gpu_queue"
    os.makedirs(LOCK_DIR, exist_ok=True)
    os.makedirs(QUEUE_DIR, exist_ok=True)

    def __init__(self, max_try=1000, sleep_duration=3, priority=5):
        self.max_try = max_try
        self.sleep_duration = sleep_duration
        self.priority = priority  # Lower number means higher priority
        self.lock_acquired = False
        self.lock_file = None
        self.gpu_id = None
        self.queue_file = None
        self.interrupted = False
        self.original_sigint_handler = None

    def __enter__(self):
        # Initialize NVML
        pynvml.nvmlInit()

        # Register signal handlers
        self.original_sigint_handler = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, self._signal_handler)

        # Get the number of GPUs
        self.num_gpus = pynvml.nvmlDeviceGetCount()
        gpu_list = list(range(self.num_gpus))
        random.shuffle(gpu_list)

        if self.num_gpus == 0:
            logger.info("No NVIDIA GPUs detected.")
            sys.exit(1)

        # Place the job in the queue with priority
        self.queue_file = os.path.join(
            self.QUEUE_DIR, f"{self.priority}_{datetime.now().timestamp()}_{uuid4()}"
        )
        open(self.queue_file, 'w').close()  # Create an empty file
        try_count = 0

        while try_count < self.max_try:
            try_count += 1

            # Check if this job is at the front of the queue
            queue_files = sorted(os.listdir(self.QUEUE_DIR))
            if os.path.basename(self.queue_file) != queue_files[0]:
                index = queue_files.index(os.path.basename(self.queue_file))
                sleep_time = min(max(self.sleep_duration, 1 * index), 100)
                logger.info(f"Current position in the queue: {index + 1}/{len(queue_files)}, sleeping for {sleep_time:.1f} seconds...")
                time.sleep(sleep_time)
                continue

            # Try to acquire a GPU starting from the random GPU
            for gpu_id in gpu_list:
                lock_path = os.path.join(self.LOCK_DIR, f"gpu_{gpu_id}.lock")
                lock = FileLock(lock_path, timeout=0)
                try:
                    lock.acquire()
                    self.lock_acquired = True
                    self.lock_file = lock
                    self.gpu_id = gpu_id
                    # Set the environment variable for the GPU
                    # os.environ["CUDA_VISIBLE_DEVICES"] = str(self.gpu_id)
                    logger.info(f"Allocated GPU {self.gpu_id}")
                    # Remove the queue file as we're proceeding
                    os.remove(self.queue_file)
                    return self  # Successfully acquired GPU
                except Timeout:
                    continue  # GPU is in use, try next one

            # No GPUs were free, wait before retrying
            logger.info(f"No GPUs are free at the moment. Retrying in {self.sleep_duration} seconds...")
            time.sleep(self.sleep_duration)

        # Failed to acquire a GPU after max tries
        self._cleanup()
        raise Exception(f"Failed to acquire a GPU after {self.max_try} attempts.")

    def __exit__(self, exc_type, exc_value, traceback):
        # Restore original signal handler
        signal.signal(signal.SIGINT, self.original_sigint_handler)

        # Release the GPU lock if acquired
        self._cleanup()

        # Shutdown NVML
        pynvml.nvmlShutdown()

    def _cleanup(self):
        # Remove the queue file
        if self.queue_file and os.path.exists(self.queue_file):
            os.remove(self.queue_file)
            logger.info("Removed queue file")

        # Release the GPU lock if acquired
        if self.lock_acquired and self.lock_file is not None:
            self.lock_file.release()
            logger.info(f"Released GPU {self.gpu_id}")
            self.lock_acquired = False
            self.lock_file = None
            self.gpu_id = None

    def _signal_handler(self, signum, frame):
        logger.info("Interrupt signal received. Cleaning up...")
        self._cleanup()
        # Restore original signal handler and re-raise the signal
        signal.signal(signal.SIGINT, self.original_sigint_handler)
        os.kill(os.getpid(), signum)



import h5py
import numpy as np
import logging
import traceback


logger = logging.getLogger(__name__)

from .aps_8idi import key as hdf_key


def put_results_in_hdf5(save_path, result, compression=None):
    """
    Save results to an HDF5 file with optional compression.

    Parameters
    ----------
    save_path : str
        Path to save the HDF5 file
    result : dict
        Dictionary of results to save
    compression : str, optional
        Compression method to use. If None, large arrays use gzip compression

    Returns
    -------
    None
    """
    def create_dataset_with_compression(f, path, data):
        """Helper function to create dataset with appropriate compression"""
        if isinstance(data, np.ndarray) and data.size > 8192 and compression is None:
            # chunks using the last two dimensions
            if data.ndim >= 2:
                    chunks = tuple([1] * (data.ndim - 2) + list(data.shape[-2:]))
            else:
                chunks = data.shape
            return f.create_dataset(path, data=data, 
                                    compression='lzf',
                                    chunks=chunks)
        return f.create_dataset(path, data=data, compression=compression)

    def save_dict_to_group(group, dictionary):
        """Helper function to save dictionary to HDF5 group"""
        for key, value in dictionary.items():
            if value is not None:
                group.create_dataset(key, data=value)

    try:
        with h5py.File(save_path, 'a') as f:
            for key, value in result.items():
                # Skip None values
                if value is None:
                    continue

                # Determine path
                path = (hdf_key["c2_prefix"] + '/' + key 
                       if key.startswith('correlation_map/') 
                       else hdf_key[key])

                # Delete existing dataset/group if it exists
                if path in f:
                    del f[path]

                # Handle dictionary values
                if isinstance(value, dict):
                    if value:  # Check if dictionary is not empty
                        group = f.create_group(path)
                        save_dict_to_group(group, value)
                # Handle non-dictionary values
                else:
                    create_dataset_with_compression(f, path, value)

    except Exception as e:
        logger.error(f"Unexpected error: {e}")
        traceback.print_exc()
        raise


if __name__ == '__main__':
    pass



import logging
import os
import sys
import h5py
import numpy as np
import torch
from typing import Optional

logger = logging.getLogger(__name__)


key_map = {
    "dqmap": "/qmap/dynamic_roi_map",
    "sqmap": "/qmap/static_roi_map",
    "mask": "/qmap/mask",
    "blemish": "/qmap/blemish",
}


def find_bin_count(qmap, minlength=None):
    count = np.bincount(qmap.ravel(), minlength=minlength)[1:]
    count = count.astype(np.int32)
    nan_idx = count == 0
    count[nan_idx] = -1
    return count, nan_idx


def average_with_index_map(
    img: torch.Tensor,
    qmap: torch.Tensor,
    size: Optional[int] = None,
    norm_factor: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """
    Computes the cluster-wise average intensity of an image tensor based on a
    quantization/segmentation map.

    Parameters
    ----------
    img : torch.Tensor
        A tensor of shape (C, H, W) or (C, N) where C is the number of images
        (or channels), and H, W represent the height and width of the image.
        The tensor contains pixel intensities.
    qmap : torch.Tensor
        A tensor of shape (H, W) or (N,) containing cluster labels
        (quantization indices) for each pixel.
    size : int, optional
        The total number of clusters (quantization levels). If not provided,
        it is inferred from `qmap`.
    norm_factor : torch.Tensor, optional
        A precomputed normalization factor for each cluster. If None, it is
        computed based on `qmap`.

    Returns
    -------
    torch.Tensor
        A tensor of shape (C, size-1) where each value represents the average
        intensity of the corresponding cluster.

    Notes
    -----
    - The function assumes `qmap` contains indices starting from 1
      (i.e., it ignores index `0`).
    - Clusters with zero pixels are handled by setting their normalization
      factor to `1` to avoid division by zero.
    """
    qmap = qmap.view(-1)
    num_pixels = qmap.shape[0]

    if size is None:
        size = qmap.max().item() + 1  # Infer size from qmap if not provided

    if norm_factor is None:
        norm_factor = torch.bincount(qmap, minlength=size)[1:]
        norm_factor[norm_factor == 0] = 1  # Avoid division by zero

    img = img.reshape(-1, num_pixels)
    summed_values = torch.stack(
        [
            torch.bincount(qmap, weights=img[n], minlength=size)[1:]
            for n in range(img.shape[0])
        ]
    )

    avg = summed_values.float() / norm_factor
    return avg


def convert_to_numpy(output_dict):
    for k, v in output_dict.items():
        if isinstance(v, torch.Tensor):
            output_dict[k] = v.float().cpu().numpy()
        if isinstance(v, np.ndarray) and v.dtype == np.float64:
            output_dict[k] = v.astype(np.float32)
    return output_dict


class XpcsQPartitionMap(object):
    def __init__(
        self,
        qmap_fname,
        flag_fix=False,
        flag_sort=False,
        dq_selection=None,
        crop_ratio_threshold=0.5,
        device="cpu",
    ) -> None:
        super().__init__()
        self.fname = qmap_fname
        self.device = device
        self.masked_ratio = 1.0
        self.masked_pixels = 0
        self.dq_dim = 0
        self.sq_dim = 0
        self.dqmap = None
        self.sqmap = None
        self.mask = None
        self.det_size = None
        self.qinfo = None
        self.flag_sort = flag_sort
        self.load(flag_fix)
        self.group_qmap(dq_selection)
        self.mask_crop = self.update_mask_crop(crop_ratio_threshold)

    def group_qmap(self, dq_selection=None):
        scount, snan_idx = find_bin_count(self.sqmap, self.sq_dim + 1)
        if self.flag_sort:
            # select and sort the qmap; good for twotime and multitau
            mask_idx_1d, qinfo = self.select_sort(dq_selection)
        else:
            # don't sort the qmap; good for multitau
            mask_idx_1d = np.where(self.mask.reshape(-1) == 1)[0]
            qinfo = None

        info_np = {
            "mask_idx_1d": mask_idx_1d,
            "sqmap_full": self.sqmap.reshape(-1),
            "dqmap_crop": self.dqmap.reshape(-1)[mask_idx_1d],
            "sqmap_crop": self.sqmap.reshape(-1)[mask_idx_1d],
            "scount": scount,
            "snan_idx": snan_idx,
        }

        info = {}
        for k, v in info_np.items():
            info[k] = torch.tensor(v).to(self.device)

        self.info_np = info_np
        self.info = info
        self.qinfo = qinfo

    def describe(self):
        logger.info(f"qmap file: {self.fname}")
        logger.info(f"flag_sort: {self.flag_sort}")
        logger.info(
            f"masked area pixels/ratio: "
            f"{self.masked_pixels}/"
            f"{self.masked_ratio * 100:.4f}%"
        )
        # omit dq = 0 which is not used
        logger.info(f"dq_map length: {np.unique(self.dqmap).size - 1}")

    def load(self, flag_fix=False):
        values = {}
        with h5py.File(self.fname, "r") as f:
            for key, real_key in key_map.items():
                if real_key in f:
                    values[key] = np.squeeze(f[real_key][()])
            # for backward compatibility; older qmap files do not have blemish
            if "blemish" not in values:
                values["blemish"] = np.copy(values["mask"])
                logger.info("no blemish data is available, using mask instead")
            else:
                logger.info("blemish data is read from qmap successfully")

        self.__dict__.update(values)
        self.mask = self.mask.astype(bool)
        self.blemish = self.blemish.astype(bool)
        print("debug", self.dqmap.shape, self.sqmap.shape)
        print(np.sum(self.mask), np.sum(self.blemish))
        self.dqmap = self.dqmap.astype(np.int32)
        self.sqmap = self.sqmap.astype(np.int32)
        if flag_fix:
            self.check_fix_qmap()
        self.det_size = self.mask.shape
        self.dq_dim = np.max(self.dqmap)
        self.sq_dim = np.max(self.sqmap)
        self.masked_pixels = int(np.sum(self.mask == 1))
        self.masked_ratio = self.masked_pixels / self.mask.size

    def update_file(self):
        logger.warning(f"update fixed qmap in file: [{self.fname}]")
        # update the qmap after fixing sqmap/dqmap inconsistency
        with h5py.File(self.fname, "a") as f:
            del f[key_map["dqmap"]]
            f[key_map["dqmap"]] = self.dqmap.astype(np.uint32)

    def update_rotation(self, det_size):
        if self.dqmap.shape != det_size:
            logger.warning("qmap is rotated 90 deg to match the detctor oriention.")
            self.dqmap = np.swapaxes(self.dqmap, 0, 1)
            self.sqmap = np.swapaxes(self.sqmap, 0, 1)
            self.mask = np.swapaxes(self.mask, 0, 1)
            self.det_size = self.mask.shape
            print(self.det_size, det_size)
            raise
            assert self.det_size == det_size, (
                "The shape of QMap does not match the raw data shape"
            )
            # need redo the preprocessing
            self.group_qmap()
            return True
        return False

    def update_mask_crop(self, crop_ratio_threshold):
        if self.masked_ratio < crop_ratio_threshold:
            logger.info(f"masked_ratio is too low. Apply crop-mask on the raw input.")
            mask_crop = self.info["mask_idx_1d"]
        else:
            mask_crop = None
        return mask_crop

    def check_fix_qmap(self):
        flag = True
        sq_list = np.sort(np.unique(self.sqmap[self.sqmap > 0]))
        for sq in sq_list:
            sq_roi = self.sqmap == sq
            dq = np.sort(np.unique(self.dqmap[sq_roi]))
            assert len(dq) in (1, 2), "only support sqmap belongs up to 2 bins"
            if len(dq) > 1:  # only works for 2
                flag = False
                idx_0 = np.logical_and(self.dqmap == dq[0], sq_roi)
                logger.warning(f"qmap inconsistent at sq: {sq} <-> {dq}: dq")
                if np.sum(idx_0) < len(sq_roi) // 2:
                    self.dqmap[idx_0] = dq[1]
                else:
                    idx_1 = np.logical_and(self.dqmap == dq[1], sq_roi)
                    self.dqmap[idx_1] = dq[0]
        return flag

    def select_sort(self, dq_selection=None):
        """
        group the sqmap and dqmap so that the pixels are continuous in memory
        """
        dqmap = self.dqmap.ravel()
        sqmap = self.sqmap.ravel()

        if dq_selection is None:
            dq_selection = np.unique(dqmap[dqmap > 0])
        dq_selection.sort()

        info = {"dq_slc": [], "sq_slc": [], "dq_idx": [], "sq_idx": [], "dq_sq_map": {}}

        q_roi = []
        dq_start = 0
        for dq in dq_selection:
            # nonzero return a tuple; (np.ndarray)
            roi = np.nonzero(dqmap == dq)[0]
            if np.sum(roi) == 0:
                continue
            sq_list = np.sort(np.unique(sqmap[roi]))
            info["dq_sq_map"][dq] = []

            sq_start = dq_start
            for sq in sq_list:
                sq_roi = np.nonzero(sqmap == sq)[0]
                if np.sum(sq_roi) == 0:
                    continue
                q_roi.append(sq_roi)
                info["sq_slc"].append(slice(sq_start, sq_start + len(sq_roi)))
                info["sq_idx"].append(sq)
                info["dq_sq_map"][dq].append(sq)
                sq_start += len(sq_roi)

            info["dq_idx"].append(dq)
            info["dq_slc"].append(slice(dq_start, dq_start + len(roi)))
            dq_start += len(roi)
            assert dq_start == sq_start, "dqmap and sqmap must be consistent"

        mask_idx_1d_sort = np.hstack(q_roi)
        dq_idx = info["dq_idx"]
        total_dq, dq_min, dq_max = len(dq_idx), min(dq_idx), max(dq_idx)
        logger.info(f"total dq: {total_dq}. min - max: {dq_min} - {dq_max}")
        return mask_idx_1d_sort, info

    def normalize_sqmap(self, img, flag_crop, apply_nan=True):
        scount = self.info["scount"]
        if flag_crop:
            sqmap = self.info["sqmap_crop"]
        else:
            sqmap = self.info["sqmap_full"]
        result = average_with_index_map(img, sqmap, self.sq_dim + 1, scount)

        if apply_nan:
            snan_idx = self.info["snan_idx"]
            result = result[..., ~snan_idx]
        # result = result.cpu().numpy()
        return result

    def recover_dimension(self, saxs2d, flag_crop, apply_mask=False):
        if flag_crop:
            pixel_num = self.det_size[0] * self.det_size[1]
            full_img = torch.zeros(pixel_num, dtype=torch.float32, device=saxs2d.device)
            full_img[self.info["mask_idx_1d"]] = saxs2d
        else:
            full_img = saxs2d

        full_img = full_img.reshape(1, *self.det_size)
        device = saxs2d.device

        if not apply_mask:
            full_img = full_img * torch.from_numpy(self.blemish).to(device)
        else:
            full_img = full_img * torch.from_numpy(self.mask).to(device)

        return full_img

    def normalize_multitau(self, res, save_G2=False):
        flag_crop = res["mask_crop"] is not None
        g2, g2_err = self.compute_g2(res["G2"], flag_crop)
        output_dir = {"delay_list": res["tau"], "g2": g2, "g2_err": g2_err}

        if save_G2:
            G2 = res["G2"].reshape(-1, res["G2"].shape[-1])
            if flag_crop:
                pixel_num = self.det_size[0] * self.det_size[1]
                value = torch.zeros(
                    G2.shape[0], pixel_num, dtype=torch.float32, device=G2.device
                )
                value[:, self.info["mask_idx_1d"]] = G2
            else:
                value = G2

            # the final shape of G2IPIF is (num_tau, 3, det_row, det_col)
            output_dir["G2IPIF"] = value.reshape(-1, 3, *self.det_size)
        output_dir = convert_to_numpy(output_dir)
        return output_dir

    def normalize_scattering(self, res):
        flag_crop = res["mask_crop"] is not None
        if res["saxs_2d"].shape == self.det_size:
            flag_crop = False
        saxs1d = self.normalize_sqmap(res["saxs_2d"], flag_crop)
        saxs1d_par = self.normalize_sqmap(res["saxs_2d_par"], flag_crop)
        saxs2d = self.recover_dimension(res["saxs_2d"], flag_crop, apply_mask=False)
        output_dir = {
            "saxs_2d": saxs2d,
            "saxs_1d": saxs1d,
            "saxs_1d_segments": saxs1d_par,
            "intensity_vs_time": res["intensity_vs_time"],
        }
        output_dir = convert_to_numpy(output_dir)
        return output_dir

    def compute_g2(self, G2, flag_crop):
        if not flag_crop:
            G2 = G2[..., self.info["mask_idx_1d"]]
        flag_crop = True

        IP_IF = G2[:, 1] * G2[:, 2]
        IP_IF[IP_IF == 0] = 1.0e8
        g2_pixel = G2[:, 0] / IP_IF

        G2_sq = self.normalize_sqmap(G2, flag_crop, apply_nan=False)
        G2_sq = G2_sq.reshape(-1, 3, G2_sq.shape[1])

        IP_IF_sq = G2_sq[:, 1] * G2_sq[:, 2]
        IP_IF_sq[IP_IF_sq == 0] = 1.0e8
        g2_sq = G2_sq[:, 0] / IP_IF_sq

        g2 = torch.zeros(G2.shape[0], self.dq_dim, device=G2.device)
        g2_err = torch.zeros_like(g2)

        dqmap = self.info["dqmap_crop"]
        sqmap = self.info["sqmap_crop"]
        for idx in range(1, self.dq_dim + 1):
            roi_dq = dqmap == idx
            if torch.sum(roi_dq) == 0:
                continue
            temp = g2_pixel[:, roi_dq]
            g2_err[:, idx - 1] = torch.std(temp, axis=1) / np.sqrt(temp.shape[1])
            # the sqmap index - 1 gives the index in the G2q
            # roi = sqmap[idx_corr].long().unique() - 1
            # [1, 2, 3, 4]
            roi_sq = sqmap[roi_dq].long().unique() - 1
            g2[:, idx - 1] = torch.mean(g2_sq[:, roi_sq], axis=1)

        return g2, g2_err


def check_and_fix_qmap(fname):
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    qpm = XpcsQPartitionMap(fname, flag_fix=False)
    flag = qpm.check_fix_qmap()
    if not flag:
        choise = input("update the qmap file? [Y/n] ")
        if choise == "Y":
            qpm.update_file()
    else:
        _, info = qpm.select_sort()
        for k, v in info["dq_sq_map"].items():
            logger.info(f"dq: {k} <-> {v}: sq")
        logger.info(f"[{fname}] is consistent")


def test():
    import time

    t0 = time.perf_counter()
    # "/scratch/xpcs_data_raw/qmap/foster202110_qmap_RubberDonut_Lq0_S180_18_D18_18.h5"
    xpm = XpcsQPartitionMap(
        "/scratch/xpcs_data_raw/qmap/qzhang20191006_qmap_Rigaku_S270_D27_log.h5",
        device="cuda:0",
    )
    print("time: ", round(time.perf_counter() - t0, 3))
    print(xpm.sqmap.shape)
    # print(np.sum(xpm.dcount))
    print(xpm.mask.dtype)
    print(xpm.masked_ratio)
    print(xpm.masked_pixels)
    import matplotlib.pyplot as plt

    plt.imshow(xpm.sqmap)
    plt.show()


if __name__ == "__main__":
    if len(sys.argv) == 2 and os.path.isfile(sys.argv[1]):
        check_and_fix_qmap(sys.argv[1])
    else:
        print("Usage: check_fix_qmap some_qmap.hdf")



import logging
import os

import time
from pathlib import Path
from typing import Any, Optional, Union

from .. import TwotimeCorrelator
from .dataset import create_dataset
from .xpcs_qpartitionmap import XpcsQPartitionMap
from .xpcs_result import XpcsResult

logger = logging.getLogger(__name__)


def solve_twotime(*args: Any, **kwargs: Any) -> Union[str, None]:
    kwargs_record = kwargs.copy()
    kwargs_record["analysis_type"] = "twotime"
    return solve_twotime_base(*args, analysis_kwargs=kwargs_record, **kwargs)


def solve_twotime_base(
    qmap: Optional[Union[str, Path]] = None,
    raw: Optional[Union[str, Path]] = None,
    output: str = "cluster_results",
    batch_size: int = 8,
    gpu_id: int = 0,
    verbose: bool = False,
    crop_ratio_threshold: float = 0.5,
    num_loaders: int = 16,
    begin_frame: int = 0,
    end_frame: int = -1,
    avg_frame: int = 1,
    stride_frame: int = 1,
    overwrite: bool = False,
    # save_G2: bool = False,
    dq_selection: Optional[Union[str, Path]] = None,
    smooth: str = "sqmap",
    analysis_kwargs: Optional[dict] = None,
    save_results: bool = True,
    **kwargs,
):

    log_level = logging.INFO if verbose else logging.ERROR
    logger.setLevel(log_level)
    device = f"cuda:{gpu_id}" if gpu_id >= 0 else "cpu"

    # create qpartitionmap
    qpm = XpcsQPartitionMap(
        qmap,
        device=device,
        flag_sort=True,
        crop_ratio_threshold=crop_ratio_threshold,
        dq_selection=dq_selection,
    )
    if verbose:
        qpm.describe()
        logger.info(f"device: {device}")

    logger.setLevel(log_level)

    # create dataset
    dset, use_loader = create_dataset(
        raw,
        device,
        mask_crop=qpm.mask_crop,
        avg_frame=avg_frame,
        begin_frame=begin_frame,
        end_frame=end_frame,
        stride_frame=stride_frame,
    )

    # in some detectors/configurations, the qmap is rotated
    qpm.update_rotation(dset.det_size)
    # determine the metadata path
    # dirname(FILES_IN_CURRENT_FOLDER) gives empty string
    meta_dir = os.path.dirname(os.path.abspath(raw))

    twotime_correlator = TwotimeCorrelator(
        qpm.qinfo,
        frame_num=dset.frame_num,
        det_size=dset.det_size,
        method="normal",
        mask_crop=qpm.mask_crop,
        window=1024,
        device=device,
    )

    logger.info("correlation solver created.")

    t_start = time.perf_counter()
    twotime_correlator.process_dataset(
        dset, verbose=verbose, use_loader=use_loader, num_workers=num_loaders
    )
    twotime_correlator.post_processing(smooth_method=smooth)
    t_end = time.perf_counter()
    t_diff = t_end - t_start
    frequency = dset.frame_num / t_diff
    logger.info(
        f"correlation finished in {t_diff:.2f}s." + f" frequency = {frequency:.2f} Hz"
    )

    t_start = time.perf_counter()
    raw_scattering = twotime_correlator.get_scattering()
    norm_scattering = qpm.normalize_scattering(raw_scattering)
    t_end = time.perf_counter()
    logger.info("normalization finished in %.3fs" % (t_end - t_start))

    # saving results to file
    if save_results:
        with XpcsResult(
            meta_dir,
            qmap,
            output,
            overwrite=overwrite,
            twotime_config=analysis_kwargs,
            rawdata_path=os.path.realpath(raw),
        ) as result_file:
            result_file.append(norm_scattering)
            for c2_payload in twotime_correlator.get_twotime_generator():
                result_file.append(c2_payload)

        logger.info(f"twotime analysis finished")
        return result_file.fname
    else:
        result_file_kwargs = {
            "meta_dir": meta_dir,
            "qmap_fname": qmap,
            "output_dir": output,
            "overwrite": overwrite,
            "twotime_config": analysis_kwargs,
        }
        return result_file_kwargs, (
            norm_scattering,
            twotime_correlator.get_twotime_generator(),
        )



from .gpu_corr_multitau import solve_multitau
from .gpu_corr_twotime import solve_twotime
from .xpcs_result import XpcsResult


def solve_corr(analysis='Both', *args, **kwargs):
    assert analysis in ['Multitau', 'Twotime', 'Both']
    if analysis == 'Multitau':
        return solve_multitau(*args, save_results=True, **kwargs)
    elif analysis == 'Twotime':
        return solve_twotime(*args, save_results=True, **kwargs)
    elif analysis == 'Both':
        rf_kwargs_m, payload_m = solve_multitau(*args, save_results=False, **kwargs)
        rf_kwargs_t, payload_t = solve_twotime(*args, save_results=False, **kwargs)
        rf_kwargs_m.update(rf_kwargs_t)
        with XpcsResult(**rf_kwargs_m) as results:
            for item in payload_m:
                results.append(item)
            for item in payload_t[1]:   # skip norm_scattering
                results.append(item)
        
        if results.success:
            return results.fname
        else:
            return None




import numpy as np
import torch
import logging
import os
import math


np_torch_map = {
    np.uint8: torch.uint8,
    np.uint16: torch.int16,
    np.int16: torch.int16,
    np.uint32: torch.int32,
}

logger = logging.getLogger(__name__)


class XpcsDataset(object):
    """ """

    def __init__(
        self,
        fname,
        begin_frame=0,
        end_frame=-1,
        stride_frame=1,
        avg_frame=1,
        batch_size=128,
        det_size=(512, 1024),
        device="cuda:0",
        use_loader=False,
        dtype=np.uint8,
        mask_crop=None,
        bin_time_s=1,  # bin time in seconds, placeholder for non-timepix detectors 
        run_config_path=None,      # run config path, placeholder for non-timepix detectors
    ):
        self.fname = fname
        self.raw_size = os.path.getsize(fname) / (1024**2)
        self.det_size = det_size
        self.pixel_num = self.det_size[0] * self.det_size[1]
        self.is_sparse = True

        self.mask_crop = mask_crop

        self.device = device

        self.batch_size = batch_size
        self.batch_num = 0
        self.frame_num = 0
        self.frame_num_raw = 0
        self.begin_frame = begin_frame
        self.end_frame = end_frame
        self.avg_frame = avg_frame
        self.stride = stride_frame

        self.use_loader = use_loader
        self.dtype = dtype
        self.dataset_type = None
        self.dtype_raw = None
        self.bin_time_s = bin_time_s

    def update_batch_info(self, frame_num: int) -> None:
        """
        Update batching settings.
        """

        def _resolve_index(idx: int, n: int) -> int:
            """Python-slice-style resolution with clamping to [0, n]."""
            if idx < 0:
                idx = n + idx
            return 0 if idx < 0 else (n if idx > n else idx)

        # --- inputs & basic checks ---
        self.frame_num_raw = frame_num
        if frame_num < 0:
            raise ValueError("frame_num must be non-negative")

        if getattr(self, "avg_frame", 1) <= 0 or getattr(self, "stride", 1) <= 0:
            raise ValueError("avg_frame and stride must be positive")

        if getattr(self, "batch_size", 1) <= 0:
            raise ValueError("batch_size must be positive")

        # note: if end_frame is <= 0, it will use all frames after begin_frame
        end = self.end_frame if self.end_frame > 0 else frame_num
        end = min(frame_num, end)

        # --- resolve indices like Python slices ---
        begin = _resolve_index(self.begin_frame, frame_num)

        if end < begin:
            raise ValueError(f"end_frame [{end}] must be >= begin_frame [{begin}]")

        # --- effective rounding to correlation block size ---
        eff_len = int(self.avg_frame) * int(self.stride)
        tot = end - begin
        assert tot > 0, "No frames to process after applying begin/end_frame."
        tot_rounded = (tot // eff_len) * eff_len
        rounded_end = begin + tot_rounded

        if rounded_end != end:
            logger.info(
                "end_frame rounded down from %d to %d to align with eff_len=%d.",
                end,
                rounded_end,
                eff_len,
            )
            end = rounded_end

        # --- finalize & batch math ---
        self.begin_frame = begin
        self.end_frame = end

        self.frame_num = (end - begin) // eff_len
        self.batch_num = (self.frame_num + self.batch_size - 1) // self.batch_size

    def update_mask_crop(self, new_mask):
        # some times the qmap's orientation is different from the dataset;
        # after rorating qmap, explicitly apply the new mask; if the original
        # mask is None (ie. not croping), then abort
        if self.mask_crop is None:
            return
        else:
            self.mask_crop = new_mask

    def update_det_size(self, det_size):
        self.det_size = tuple(det_size)
        self.pixel_num = self.det_size[0] * self.det_size[1]

    def get_raw_index(self, idx):
        # get the raw index list
        eff_len = self.stride * self.avg_frame * self.batch_size
        beg = self.begin_frame + eff_len * idx
        end = min(self.end_frame, beg + eff_len)
        size = (end - beg) // self.stride
        return beg, end, size

    def get_sparsity(self):
        x = self.__getitem__(0)[0]
        y = (x > 0).sum() / self.pixel_num
        self.dtype_raw = x.dtype
        self.__reset__()
        return y

    def get_description(self):
        result = {}
        for key in [
            "fname",
            "frame_num_raw",
            "is_sparse",
            "frame_num",
            "det_size",
            "device",
            "batch_size",
            "batch_num",
            "dataset_type",
        ]:
            result[key] = self.__dict__[key]
        result["frame_info"] = (
            "(begin, end, stride, avg) = ("
            f"{self.begin_frame}, {self.end_frame}, "
            f"{self.stride}, {self.avg_frame})"
        )
        return result

    def describe(self):
        for key, val in self.get_description().items():
            logger.info(f"{key}: {val}")

        if self.mask_crop is not None:
            valid_size = self.mask_crop.shape[0]
        else:
            valid_size = self.pixel_num
        logger.info(f"dtype: {self.dtype}")
        logger.info(f"valid_size: {valid_size}")
        logger.info(f"sparsity: {self.get_sparsity():.4f}")
        logger.info(f"raw dataset file size: {self.raw_size:.2f} MB")

    def __len__(self):
        return self.batch_num

    def __reset__(self):
        return

    def __getbatch__(self, idx):
        raise NotImplementedError

    def __getitem__(self, idx):
        x = self.__getbatch__(idx)
        if self.avg_frame > 1:
            x = x.reshape(-1, self.avg_frame, x.shape[-1])
            x = torch.mean(x.float(), axis=1)
        return x

    def to_rigaku_bin(self, fname):
        """
        convert to rigaku binary format
        """
        offset = 0
        fid = open(fname, "a")
        for n in range(self.batch_num):
            x = self.__getitem__(n)
            idx = np.nonzero(x)
            count = x[idx]
            frame = idx[0] + offset
            index = idx[1].astype(np.int64)
            offset += x.shape[0]

            y = np.zeros_like(count, dtype=np.int64)
            y += frame.astype(np.int64) << 40
            y += index << 16
            y = y + count
            y.tofile(fid, sep="")

        fid.close()

    def sparse_to_dense(self, index, frame, count, size):
        if isinstance(index, np.ndarray):
            # using numpy array; will be sent to device by dataloader
            x = np.zeros((size, self.pixel_num), dtype=self.dtype)
            x[frame, index] = count
        else:
            x = torch.zeros(
                (size, self.pixel_num),
                dtype=np_torch_map[self.dtype],
                device=index.device,
            )
            x[frame.long(), index.long()] = count

        return x



import numpy as np
import torch
import os
import logging
import time

from tqdm import tqdm, trange
from .help_functions import gen_tau_bin, sort_tau_bin, is_power_two
from torch.utils.data import DataLoader
# import matplotlib.pyplot as plt
# import skimage.io as skio

logger = logging.getLogger(__name__)


def compute_dtype(max_count, levels):
    # get the dtype that can hold the accumulation of photons without losing
    # precision
    dtype_list = [
        [1 / (2**7 - 1), torch.bfloat16, 2],  # bfloat16 has  bits
        [1 / (2**23 - 1), torch.float32, 4],  # float32 has 23 bits
        [1 / (2**52 - 1), torch.float64, 8],  # float64
    ]

    dtype_ret = []
    for level in levels:
        accu_count = max_count * (2**level)
        min_percent = 1.0 / accu_count
        for n in range(len(dtype_list)):
            if min_percent > dtype_list[n][0]:
                dtype_ret.append(dtype_list[n][1:])
                break

    return dtype_ret


def compute_queue(levels_num, queue_size, queue_level):
    x = np.arange(levels_num, dtype=np.int64)
    y = queue_size // (2 ** (x // queue_level))
    y[y < 8] = 8
    return y


class MultitauCorrelator(object):
    def __init__(
        self,
        det_size,
        frame_num,
        device="cpu",
        queue_size: int = 64,
        queue_level: int = 4,
        auto_queue=True,
        mask_crop=None,
        max_memory=36.0,
        normalize_frame=False,
        max_count=7,
        num_partial_g2=0,
        qpm=None,
    ) -> None:
        self.det_size = det_size
        self.is_sparse = False
        self.max_memory = max_memory

        self.mask_crop = mask_crop
        if self.mask_crop is not None:
            self.pixel_num = mask_crop.shape[0]
        else:
            self.pixel_num = det_size[0] * det_size[1]

        self.frame_num = frame_num
        self.normalize_frame = normalize_frame

        assert isinstance(queue_size, int), "queue_size must be an integer"
        assert is_power_two(queue_size), "queue_size must be a power of 2"
        self.queue_size = queue_size

        tau_bin = gen_tau_bin(self.frame_num)
        self.tau_bin = tau_bin

        tau_max, levels, tau_in_level = sort_tau_bin(tau_bin, frame_num)
        self.levels_num = len(levels)
        self.tau_num = tau_bin.shape[1]
        levels_num = len(levels)

        self.tau_rev = tau_max
        self.dtype_list = compute_dtype(max_count, levels)

        # self.device = torch.device(device_type)
        self.device = device

        self.ct = [None for n in range(levels_num)]
        self.tau_all = []
        self.tau_in_level = tau_in_level

        # over write queue_size and queue_level
        if auto_queue:
            queue_size, queue_level = self.auto_set_queue(queue_size)
        queue = compute_queue(self.levels_num, queue_size, queue_level)
        self.queue_info = (auto_queue, queue_size, queue_level)

        for n in range(levels_num):
            curr_tau_all = self.tau_rev + queue[n]
            self.ct[n] = torch.zeros(
                size=(curr_tau_all, self.pixel_num),
                dtype=self.dtype_list[n][0],
                device=self.device,
            )
            self.tau_all.append(curr_tau_all)
            # print(self.dtype_list[n])

        # account for the overflow photons
        self.ct_overflow = torch.zeros(size=(self.pixel_num,), device=self.device)

        # G2, IP and IF
        self.g2 = torch.zeros(
            size=(self.tau_num, 3, self.pixel_num),
            dtype=torch.float32,
            device=self.device,
        )
        # pointer to the data enty in each level
        self.pt = [self.tau_rev] * levels_num

        self.num_partial_g2 = num_partial_g2
        if self.num_partial_g2 > 0:
            assert qpm is not None, (
                "qpm must be provided when compute_partial_g2 is True"
            )
            self.act_length = np.zeros(self.tau_num)
            self.qpm = qpm
            self.act_g2 = torch.zeros_like(self.g2)

        # advanced flag
        self.ad = [False] * levels_num

        self.par_level = 0
        while self.frame_num // (2**self.par_level) > 16:
            self.par_level += 1

        # self.intt = torch.zeros(size=(frame_num, 2), dtype=torch.float32)
        self.intt = []
        self.saxs_2d = torch.zeros_like(self.ct_overflow)
        self.saxs_2d_par = []
        self.g2_partial = []

        self.levels = levels
        self.tau_in_level = tau_in_level
        self.current_frame = 0

    def describe(self):
        logger.info(f"queue information (auto, size, level): {self.queue_info}")
        dlist = [str(self.dtype_list[n][0]) for n in range(len(self.dtype_list))]
        dlist_uniq = list(set(dlist))
        dlist_uniq.sort()
        for x in dlist_uniq:
            logger.info(f"stage @ {dlist.count(x):02d}: {x}")
        logger.info(f"max_memory:  {self.max_memory} GB")
        logger.info(
            f"normalize frame to account for intensity fluctuation: {self.normalize_frame}"
        )

    def reset(self):
        for n in range(self.levels_num):
            self.ct[n].zero_()

        # account for the overflow photons
        self.ct_overflow = 0

        # G2, IP and IF
        self.g2.zero_()
        # pointers for each level
        self.pt = [self.tau_rev] * self.levels_num

        # advanced flag
        self.ad = [False] * self.levels_num
        self.intt = []
        self.saxs_2d = None
        self.saxs_2d_par = []
        self.current_frame = 0

    def auto_set_queue(self, min_queue_size, overhead=0.25):
        # the size of each frame in all levels;
        sz = np.zeros(self.levels_num)
        for n in range(self.levels_num):
            sz[n] = self.dtype_list[n][1] * self.pixel_num / (1024**3)
        # RAM reservation for the G2/IP/IF array
        g2_sz = 4 * self.pixel_num / (1024**3) * self.tau_num * 3

        max_memory = self.max_memory / (1 + overhead)
        max_memory -= np.sum(sz) * self.tau_rev + g2_sz

        # choose a starting level not equal to 1 so the flush of different
        # levels don't happen at the same time.
        queue_level = 3
        queue_size = min_queue_size

        def get_size(*args):
            return np.sum(compute_queue(*args) * sz)

        while get_size(self.levels_num, queue_size * 2, queue_level) <= max_memory:
            queue_size *= 2
            if queue_size > self.frame_num:
                break

        while get_size(self.levels_num, queue_size, queue_level + 1) <= max_memory:
            queue_level += 1
            if queue_level >= self.levels_num:
                break

        return queue_size, queue_level

    def process_numpy(self, *args):
        if not self.is_sparse:
            y = torch.from_numpy(args[0]).to(self.device)
            self.__process_dense__(y)
        else:
            index, frame, count, size = args
            index = torch.from_numpy(index).to(self.device)
            frame = torch.from_numpy(frame).to(self.device)
            count = torch.from_numpy(count).to(self.device)
            self.__process_sparse__(index, frame, count, size)

    def process(self, *args):
        if not self.is_sparse:
            self.__process_dense__(*args)
        else:
            self.__process_sparse__(*args)

    def __process_sparse__(self, index, frame, count, size):
        x = torch.zeros((size, self.pixel_num), count.dtype, device=self.device)
        x[frame.long(), index.long()] = count
        self.__process_dense__(x)

    def __process_dense__(self, x=None):
        # convert raw input (likely uint16) to float
        # prefer to pass an input with size of 1/N queue size
        # x = x.float()
        if self.num_partial_g2 > 0:
            prev_seg = self.current_frame // (self.frame_num // self.num_partial_g2)
            self.current_frame += x.shape[0]
            next_seg = self.current_frame // (self.frame_num // self.num_partial_g2)
            flag_g2_part = next_seg > prev_seg

            if flag_g2_part:
                g2_part_2d = self.calculate_partial_g2()
                g2_part_dict = self.qpm.normalize_multitau(g2_part_2d, save_G2=False)
                g2_part_dict["num_frames"] = self.act_length[0]
                self.g2_partial.append(g2_part_dict)
        else:
            self.current_frame += x.shape[0]
            flag_g2_part = False

        last_flag = self.current_frame == self.frame_num

        for level in self.levels:
            if x is not None:
                pt0 = self.pt[level]
                self.ct[level][pt0 : pt0 + x.shape[0]] = x
                self.pt[level] += x.shape[0]

            if self.pt[level] != self.tau_all[level] and not last_flag:
                # most likely case
                return
            else:
                x = self.flush_level(level, last_flag)

                # accumulate the overflow photons; most likely to be zero due
                # to the tau/fold list setting; last level; must return
                if level == self.levels[-1] and x is not None:
                    self.ct_overflow += torch.sum(x, dim=0)
                    return

                # if there are no left events and it's not the final batch
                if x is None and not last_flag:
                    return

    def flush_level(self, level, last_flag):
        beg = self.tau_rev
        end = self.pt[level]
        # when the channel has nothing;
        if end - beg <= 0:
            return None

        if level == 0:
            # sl = slice(self.current_frame - end + beg, self.current_frame)
            intt = torch.mean(self.ct[level][beg:end].float(), dim=1)
            self.intt.append(intt)
            self.saxs_2d += torch.sum(self.ct[level][beg:end].float(), dim=0)
            if self.normalize_frame:
                self.ct[level][beg:end] /= intt.reshape(end - beg, 1)

        for tau, tid, _ in self.tau_in_level[level]:
            sl1 = slice(beg - tau, end - tau)
            sl2 = slice(beg, end)
            xy = torch.sum(self.ct[level][sl1] * self.ct[level][sl2], dim=0)
            self.g2[tid, 0] += xy

            if self.num_partial_g2 > 0:
                if self.act_length[tid] == 0:
                    self.act_length[tid] = end - beg - tau
                else:
                    self.act_length[tid] += end - beg
                self.act_g2[tid, 0] += xy
                self.act_g2[tid, 1] += torch.sum(self.ct[level][sl1], dim=0)

        # deal with IP; it may happen when the channel advances or it's the
        # last batch and needs flush;
        if not self.ad[level]:
            for tau, tid, _ in self.tau_in_level[level]:
                sl = slice(beg, beg + tau)
                self.g2[tid, 1] -= torch.sum(self.ct[level][sl], dim=0)
            self.ad[level] = True

        # deal with IF; it only happends for the final batch
        if last_flag:
            for tau, tid, _ in self.tau_in_level[level]:
                sl = slice(end - tau, end)
                self.g2[tid, 2] -= torch.sum(self.ct[level][sl], dim=0)

        if level == self.par_level:
            self.saxs_2d_par.append(self.ct[level][beg:end].clone().detach())

        # the number of elements should be an even number for all; but the
        # last call can be an odd number; advance the event; the left element
        # will be used to calcuate the sum
        avg_len = (end - beg) // 2 * 2
        self.pt[level] -= avg_len
        avg_sl = slice(beg, beg + avg_len)

        x = torch.sum(
            self.ct[level][avg_sl].view(avg_len // 2, 2, self.pixel_num), dim=1
        )

        # copy the ending to buffer
        tmp = self.ct[level][avg_len:end].clone().detach()
        self.ct[level][0 : end - avg_len] = tmp

        return x

    def get_partial_g2(self):
        if self.num_partial_g2 <= 0:
            return {"g2_partial": None}

        unique_g2s = []
        for g2_part_dict in self.g2_partial:
            if len(unique_g2s) > 0 and np.allclose(
                g2_part_dict["g2"], unique_g2s[-1]["g2"]
            ):
                continue
            else:
                unique_g2s.append(g2_part_dict)
        g2 = np.stack([g2_part["g2"] for g2_part in unique_g2s], axis=0)
        g2[g2 == 0] = np.nan

        g2_err = np.stack([g2_part["g2_err"] for g2_part in unique_g2s], axis=0)
        labels = np.array([g2_part["num_frames"] for g2_part in unique_g2s])
        result = {
            "g2_partial": g2,
            "g2_partial_err": g2_err,
            "g2_partial_labels": labels,
        }
        return result

    def calculate_partial_g2(self):
        # rescale the average and add total flux to IP and IF
        if self.num_partial_g2 <= 0:
            raise ValueError("compute_partial_g2 must be True to get partial G2")

        g2_part = self.act_g2.clone().detach()
        for level in self.levels:
            for _, tid, _ in self.tau_in_level[level]:
                eff_length = self.act_length[tid]
                g2_part[tid, 0] /= eff_length * 2 ** (level * 2)
                g2_part[tid, 1] /= eff_length * 2**level
                g2_part[tid, 2] = g2_part[tid, 1]

        output_multitau = {
            "G2": g2_part.float(),
            "mask_crop": self.mask_crop,
            "tau": self.tau_bin[0, :],
        }
        return output_multitau

    def post_process(self):
        # rescale the average and add total flux to IP and IF
        tot = self.ct_overflow
        for level in np.flip(self.levels):
            sl = slice(self.tau_rev, self.pt[level])
            tot = tot + torch.sum(self.ct[level][sl], dim=0)
            # normalization
            # G2 = (x / scl) * (y / scl) / eff_len  -> eff_len * scl ^ 2
            # IP = (x / scl)             / eff_len  -> eff_len * scl
            # IF = (y / scl)             / eff_len  -> eff_len * scl
            for _, tid, eff_length in self.tau_in_level[level]:
                self.g2[tid, 0] /= eff_length * 2 ** (level * 2)
                self.g2[tid, 1:3] += tot
                self.g2[tid, 1:3] /= eff_length * 2**level

        # get saxs2d
        # self.saxs_2d = torch.unsqueeze(tot / self.frame_num, 0)
        self.saxs_2d /= self.frame_num
        if len(self.saxs_2d_par) > 0:
            self.saxs_2d_par = torch.vstack(self.saxs_2d_par) / self.frame_num
        else:
            self.saxs_2d_par = torch.unsqueeze(self.saxs_2d, 0)
        return

    def process_dataset(self, ds, verbose=True, use_loader=False, num_workers=16):
        if not use_loader:
            xrange = trange if verbose else range
            for n in xrange(len(ds)):
                self.process(ds[n])
        else:
            # only pin memory if using GPU for computing
            pin_memory = self.device != "cpu"
            dl = DataLoader(
                ds,
                batch_size=None,
                pin_memory=pin_memory,
                num_workers=num_workers,
                prefetch_factor=4,
            )
            logger.info(f"using {num_workers} workers to load data")

            if verbose:
                container = tqdm(dl, total=len(ds), desc="progress")
            else:
                container = dl

            for x in container:
                x = x.to(self.device, non_blocking=True)
                self.process(x)

        self.post_process()

    def get_results(self):
        intt = torch.hstack(self.intt).float()
        tline = torch.arange(intt.shape[0], device=intt.device)
        intt = torch.vstack([tline, intt])

        output_scattering = {
            "saxs_2d": self.saxs_2d.float(),
            "saxs_2d_par": self.saxs_2d_par.float(),
            "intensity_vs_time": intt.float(),
            "mask_crop": self.mask_crop,
        }

        output_multitau = {
            "G2": self.g2.float(),
            "mask_crop": self.mask_crop,
            "tau": self.tau_bin[0, :],
        }
        return output_scattering, output_multitau


def read_data(
    det_size, fname="../xpcs_data_simulation/simulation_0010k_sparse_0.005.bin"
):
    det_size_1d = det_size[0] * det_size[1]
    data = np.fromfile(fname, dtype=np.uint16)
    data = data.astype(np.int16)
    data = data.reshape(-1, det_size_1d)
    return data, data.shape[0]


def example(
    queue_size=512,
    fname="simulation_0010k_sparse_0.005.hdf",
    data_dir="../xpcs_data_simulation",
):
    path = os.path.join(data_dir, fname)
    # detector = XpcsDetector(path)
    # data, frame_num = read_data(det_size=(512, 512))
    # data = torch.tensor(data, device='cuda:0')
    batch_size = queue_size
    # det_size = (1024, 512)
    det_size = (1813, 1558)
    frame_num = 100000

    logger.info("frame_num = %d", frame_num)
    logger.info("queue_size = %d", queue_size)
    logger.info("det_size = %s", det_size)
    xb = MultitauCorrelator(
        det_size=det_size, frame_num=frame_num, queue_size=queue_size, device="cuda:1"
    )
    xb.debug()
    stime = time.perf_counter()
    for n in tqdm(range(frame_num // batch_size + 1), colour="green"):
        sz = min(frame_num, (n + 1) * batch_size) - batch_size * n
        x = torch.ones(
            (sz, det_size[0] * det_size[1]), device=xb.device, dtype=torch.bfloat16
        )
        xb.process(x)

    etime = time.perf_counter()
    logger.info("processing frequency is %.4f" % (frame_num / (etime - stime)))

    xb.post_process()
    return


def test_dtype():
    levels = list(np.arange(32))
    max_count = 1
    compute_dtype(max_count, levels)


if __name__ == "__main__":
    # for queue_size in 2 ** np.arange(3, 15):
    #     example(queue_size=queue_size)
    example(queue_size=512)
    # test_dtype()



import numpy as np
import logging
import torch
from .xpcs_dataset import XpcsDataset
from timepix_dataset.dataset import TimepixRawDataset


logger = logging.getLogger(__name__)


class TimepixAnalysisDataset(XpcsDataset):
    def __init__(
        self,
        *args,
        dtype=np.uint8,
        total_frames=None,
        bin_time_s=1e-6,
        run_config_path=None,
        **kwargs,
    ):
        super(TimepixAnalysisDataset, self).__init__(*args, dtype=dtype, **kwargs)
        self.dataset_type = "Timepix4Dataset"
        self.is_sparse = True
        self.dtype = np.uint8
        self.raw_dataset = TimepixRawDataset(self.fname, run_config_path)
        self.update_det_size(self.raw_dataset.det_size)

        self.ifc_list, self.mem_addr_list = self.read_data(
            total_frames, bin_time_s=bin_time_s
        )
        self.ifc_list = self.to_device()

    def to_device(self, max_fsize_gb_in_gpu=1.0):
        ifc_device = []
        for ifc in self.ifc_list:
            module_ifc = self.to_device_single_module(ifc, max_fsize_gb_in_gpu)
            ifc_device.append(module_ifc)
        return ifc_device

    def to_device_single_module(self, ifc, max_fsize_gb_in_gpu=1.0):
        total_size = sum([arr.nbytes for arr in ifc]) / 1024**3
        if total_size <= max_fsize_gb_in_gpu and self.device.startswith("cuda"):
            logger.info("Preloading timepix rawdata to GPU memory")
            device = self.device
        else:
            device = "cpu"
            logger.info(f"Total data size {total_size:.2f} GB, offloading to CPU RAM")
        index = torch.tensor(ifc[0], device=device)
        frame = torch.tensor(ifc[1], device=device)
        count = torch.tensor(ifc[2], device=device)
        return (index, frame, count)

    def read_data(self, total_frames=None, bin_time_s=1e-6):
        ifc_list = self.raw_dataset.apply_time_binning(bin_time_s)

        # the last frame is dropped to avoid partial frames
        num_frames = max([frame[-1] for _, frame, _ in ifc_list])
        # update frame_num and batch_num
        self.update_batch_info(num_frames)

        all_index = np.arange(0, self.frame_num_raw + 2)
        all_index[-1] = self.frame_num_raw

        mem_addr_list = []
        for ifc in ifc_list:
            # build an index map for all indexes
            beg = np.searchsorted(ifc[1], all_index[:-1])
            end = np.searchsorted(ifc[1], all_index[1:])
            mem_addr = [slice(a, b) for a, b in zip(beg, end)]
            mem_addr_list.append(mem_addr)
        return ifc_list, mem_addr_list

    def __getbatch_single_module__(self, idx, module_index=0):
        # frame begin and end
        beg, end, size = self.get_raw_index(idx)
        ifc = self.ifc_list[module_index]
        device = ifc[0].device

        pixel_num = self.raw_dataset.mod_size[0] * self.raw_dataset.mod_size[1]
        x = torch.zeros(size, pixel_num, dtype=torch.bfloat16, device=device)
        if self.stride == 1:
            # the data is continuous in RAM; convert by batch
            sla, slb = (
                self.mem_addr_list[module_index][beg],
                self.mem_addr_list[module_index][end],
            )
            a, b = sla.start, slb.start
            x[ifc[1][a:b].long() - beg, ifc[0][a:b].long()] = ifc[2][a:b].to(
                torch.bfloat16
            )
        else:
            # the data is discrete in RAM; convert frame by frame
            for n, idx in enumerate(np.arange(beg, end, self.stride)):
                sl = self.mem_addr_list[idx]
                x[n, ifc[0][sl].long()] = ifc[2][sl].to(torch.bfloat16)

        x = x.to(self.device)
        return x

    def __getbatch__(self, idx):
        # deal with with single modules
        if self.raw_dataset.num_chips == 1:
            canvas = self.__getbatch_single_module__(idx, module_index=0)
        # deal with with multiple modules
        else:
            canvas = torch.zeros(
                self.batch_size,
                *self.raw_dataset.det_size,
                dtype=torch.uint8,
                device=self.device,
            )
            for module_index in range(self.raw_dataset.num_chips):
                _data = self.__getbatch_single_module__(idx, module_index=module_index)
                actual_batch_size = _data.shape[0]

                _data = _data.to(self.device)
                _data = _data.view(actual_batch_size, *self.raw_dataset.mod_size)
                layout = self.raw_dataset.layout[module_index]
                canvas[:actual_batch_size, :, layout] = _data

            canvas = canvas[0:actual_batch_size].view(actual_batch_size, self.pixel_num)

        if self.mask_crop is not None:
            canvas = canvas[:, self.mask_crop]

        return canvas



import logging
import time
import torch
from boost_corr import MultitauCorrelator 


logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s T+%(relativeCreated)05dms [%(filename)s]: %(message)s",
    datefmt="%m-%d %H:%M:%S")

logger = logging.getLogger(__name__)


device = 'cuda:0'
    

def test(queue_size=128):
    batch_size = queue_size
    # det_size = (1024, 512)
    det_size = (516, 1556)
    frame_num = 100 * 1000

    logger.info('frame_num = %d', frame_num)
    logger.info('queue_size = %d', queue_size)
    logger.info('det_size = %s', det_size)

    xb = MultitauCorrelator(det_size=det_size,
                   frame_num=frame_num,
                   queue_size=512,
                   device='cuda:1')
    # xb.debug()
    stime = time.perf_counter()
    for n in range(frame_num // batch_size + 1):
        sz = min(frame_num, (n + 1) * batch_size) - batch_size * n
        x = torch.ones((sz, det_size[0] * det_size[1]),
                       device=xb.device,
                       dtype=torch.bfloat16)
        xb.process(x)

    etime = time.perf_counter()
    logger.info('processing frequency is %.4f' % (frame_num / (etime - stime)))

    # xb.post_process()
    return


def test_host_gpu(queue_size=1):
    batch_size = queue_size
    # det_size = (1024, 512)
    # det_size = (516, 1556)
    det_size = (2048, 2048)
    frame_num = 100 * 1000

    logger.info('frame_num = %d', frame_num)
    logger.info('queue_size = %d', queue_size)
    logger.info('det_size = %s', det_size)


    stime = time.perf_counter()
    for n in range(frame_num // batch_size + 1):
        sz = min(frame_num, (n + 1) * batch_size) - batch_size * n
        # x = torch.ones((sz, det_size[0] * det_size[1]), device='cpu',
        #         dtype=torch.uint8, pin_memory=True)
        x = torch.ones((sz, det_size[0] * det_size[1]), device='cpu',
                dtype=torch.uint8, pin_memory=True)
        x = x.to(device)

    etime = time.perf_counter()
    logger.info('processing frequency is %.4f' % (frame_num / (etime - stime)))


if __name__ == '__main__':
    test_host_gpu()




import os
import h5py
from .imm_handler import ImmDataset
from .rigaku_handler import RigakuDataset
from .rigaku_3M_handler import Rigaku3MDataset
from .hdf_handler import HdfDataset


def create_dataset(
    raw_fname,
    device,
    mask_crop,
    avg_frame,
    begin_frame,
    end_frame,
    stride_frame,
    bin_time_s=1e-6,
    run_config_path=None,
):
    if not os.path.isfile(raw_fname):
        raise FileNotFoundError(f"The raw_file '{raw_fname}' does not exist.")

    ext = os.path.splitext(raw_fname)[-1]
    # use_loader is set for HDF files, it can use multiple processes to read
    # large HDF file;
    use_loader = False
    if ext == ".bin":
        dataset_method = RigakuDataset
        use_loader = False
        batch_size = 1024
    elif raw_fname.endswith(".bin.000"):
        dataset_method = Rigaku3MDataset
        use_loader = False
        batch_size = 256
    elif ext in [".imm", ".h5", ".hdf", ".hdf5"]:
        # Check if it's an HDF5 file by trying to detect HDF5 format
        # If not HDF5, treat it as IMM format
        real_raw = os.path.realpath(raw_fname)
        try:
            is_hdf5 = h5py.is_hdf5(real_raw)
        except (OSError, IOError):
            # File might be corrupted or inaccessible
            is_hdf5 = False
        
        if is_hdf5:
            dataset_method = HdfDataset
            use_loader = True
            batch_size = 8
        else:
            dataset_method = ImmDataset
            use_loader = False
            batch_size = 8
    elif raw_fname.endswith((".tpx", ".tpx.000", ".tpx.001", ".tpx.002")):
        from .timepix4_handler import TimepixAnalysisDataset

        dataset_method = TimepixAnalysisDataset
        use_loader = False
        batch_size = 1024
    else:
        raise TypeError(f"File type [{ext}] is not supported")

    dset = dataset_method(
        raw_fname,
        batch_size=batch_size,
        device=device,
        mask_crop=mask_crop,
        avg_frame=avg_frame,
        begin_frame=begin_frame,
        end_frame=end_frame,
        stride_frame=stride_frame,
        bin_time_s=bin_time_s,
        run_config_path=run_config_path,
    )

    return dset, use_loader



import numpy as np
from .xpcs_dataset import XpcsDataset
from .rigaku_handler import RigakuDataset
import os
import logging
import torch
from .help_functions import get_number_of_frames_from_binfile
import time


logger = logging.getLogger(__name__)


class Rigaku3MDataset(XpcsDataset):
    """
    Parameters
    ----------
    filename: string
        path to Rigaku3M binary.000 file
    """

    def __init__(self, *args, dtype=np.uint8, gap=(70, 52), layout=(3, 2), **kwargs):
        super(Rigaku3MDataset, self).__init__(*args, dtype=dtype, **kwargs)
        self.dataset_type = "Rigaku 64bit Binary"
        self.is_sparse = True
        self.dtype = np.uint8
        self.container = self.get_modules(dtype, **kwargs)
        self.gap = gap
        self.layout = layout
        self.update_info()

    def get_modules(self, dtype, timeout=10, **kwargs):
        # this is the order for the 6 modules
        index_list = (5, 0, 4, 1, 3, 2)
        flist = [self.fname[:-3] + f"00{n}" for n in index_list]
        for f in flist:
            t0 = timeout
            while not os.path.isfile(f) and t0 > 0:
                time.sleep(1)
                t0 -= 1
            if t0 <= 0:
                raise FileNotFoundError(f"timeout waiting for file {f}")
            assert os.path.isfile(f), "file not found: {f}"
        # determine the total number of frames
        total_frames = [get_number_of_frames_from_binfile(f) for f in flist]
        for n, f in enumerate(flist):
            logger.info(f"{f}: {total_frames[n]}")
        total_frames = max(total_frames)

        kwargs.pop("mask_crop", None)
        return [
            RigakuDataset(
                f, dtype=dtype, mask_crop=None, total_frames=total_frames, **kwargs
            )
            for f in flist
        ]

    def update_info(self):
        frame_num = [x.frame_num_raw for x in self.container]
        assert len(list(set(frame_num))) == 1, "frame number mismatch in the 6 modules"
        self.update_batch_info(frame_num[0])
        shape_one = self.container[0].det_size
        shape = [
            shape_one[n] * self.layout[n] + self.gap[n] * (self.layout[n] - 1)
            for n in range(2)
        ]
        self.det_size = tuple(shape)
        self.det_size_one = shape_one

    def append_data(self, canvas, index, data_module):
        row = index // 2
        col = index % 2
        st_v = row * (self.det_size_one[0] + self.gap[0])
        sl_v = slice(st_v, st_v + self.det_size_one[0])
        st_h = col * (self.det_size_one[1] + self.gap[1])
        sl_h = slice(st_h, st_h + self.det_size_one[1])
        canvas[:, sl_v, sl_h] = data_module.reshape(-1, *self.det_size_one)

        return canvas

    def __getbatch__(self, idx):
        _, _, size = self.get_raw_index(idx)
        canvas = torch.zeros(
            size, *self.det_size, dtype=torch.uint8, device=self.device
        )
        for index, module in enumerate(self.container):
            temp = module.__getbatch__(idx)
            canvas = self.append_data(canvas, index, temp)

        canvas = canvas.reshape(size, -1)
        if self.mask_crop is not None:
            canvas = canvas[:, self.mask_crop]

        return canvas


def test():
    fname = "../../../tests/data/verify_circular_correlation/F091_D100_Capillary_Post_att00_Lq0_Rq0_00001/F091_D100_Capillary_Post_att00_Lq0_Rq0_00001.bin"
    ds = RigakuDataset(fname)


if __name__ == "__main__":
    test()



#!/usr/bin/env python

"""Tests for multitau correlation."""


import unittest
from boost_corr import MultitauCorrelator 
from boost_corr.help_functions import gen_tau_bin
import torch
import numpy as np


def run_and_get_result(det_size=(5, 7), frame_num=1024, device='cpu'):
    num_elm = det_size[0] * det_size[1]
    mc = MultitauCorrelator(det_size, frame_num, device=device,
                            max_count=4096)
    # use high max_count to make sure all levels are float32
    for _ in range(frame_num):
        x = torch.ones(det_size, device=device)
        mc.process(x.reshape(-1, num_elm))
    mc.post_process()
    result = mc.get_results()
    return result


class TestBoost_corr(unittest.TestCase):
    """Tests for `boost_corr` package."""

    def test_000_cpu(self):
        frame_num = 1024
        result = run_and_get_result(frame_num=frame_num, device='cpu')
        intt = result['intt'][1]
        intt_real = torch.ones_like(intt)
        self.assertEqual(torch.allclose(intt, intt_real), True)

        saxs2d = result['saxs2d']
        saxs2d_real = torch.ones_like(saxs2d)
        self.assertEqual(torch.allclose(saxs2d, saxs2d_real), True)

        tau = result['tau']
        tau_bin = gen_tau_bin(frame_num)
        self.assertEqual(np.all(tau == tau_bin[0]), True)
        
        g2 = result['g2']
        g2_real = torch.ones_like(result['g2'])
        self.assertEqual(torch.allclose(g2, g2_real), True)

    def test_001_gpu(self):
        frame_num = 1024
        result = run_and_get_result(frame_num=frame_num, device='cuda:0')
        intt = result['intt'][1]
        intt_real = torch.ones_like(intt)
        self.assertEqual(torch.allclose(intt, intt_real), True)

        saxs2d = result['saxs2d']
        saxs2d_real = torch.ones_like(saxs2d)
        self.assertEqual(torch.allclose(saxs2d, saxs2d_real), True)

        tau = result['tau']
        tau_bin = gen_tau_bin(frame_num)
        self.assertEqual(np.all(tau == tau_bin[0]), True)
        
        g2 = result['g2']
        g2_real = torch.ones_like(result['g2'])
        self.assertEqual(torch.allclose(g2, g2_real), True)


if __name__ == '__main__':
    unittest.main()



import os
import sys
import shutil
import h5py


def append_metadata_qmap(output_fname, meta_fname=None, qmap_fname=None,
                         meta_type='nexus'):
    """
    Appends metadata to an output file.
    Parameters
    ----------
    output_fname : str
        The name of the output file.
    meta_fname : str, optional
        The name of the metadata file.
    qmap_fname : str, optional
        The name of the qmap file.
    meta_type : str, optional
        The type of metadata. Default is 'nexus'.
    """
    assert meta_type in ('nexus')
    # copy the metadata file
    shutil.copyfile(meta_fname, output_fname)
    os.chmod(output_fname, 0o644)
    # copy_qmap(output_fname, qmap_fname)
    copy_qmap(source_file=qmap_fname, target_file=output_fname)


def copy_qmap(source_file, target_file, source_group='/qmap',
              target_group='/xpcs/qmap'):
    """
    Copies a group from one HDF5 file to another with the option to rename the group.

    Parameters:
    -----------
        source_file (str): Path to the source HDF5 file.
        target_file (str): Path to the target HDF5 file.
        source_group (str): Name of the source group to copy.
        target_group (str): Name of the target group in the target file.

    Returns:
    ----------
        None
    """
    with h5py.File(source_file, "r") as src, h5py.File(target_file, "a") as tgt:
        src.copy(source_group, tgt, name=target_group)



if __name__ == '__main__':
    append_metadata_qmap(sys.argv[0], sys.argv[1], sys.argv[2])



import logging
import os
import time
from pathlib import Path
from typing import Any, Optional, Union

from .. import MultitauCorrelator
from .dataset import create_dataset
from .xpcs_qpartitionmap import XpcsQPartitionMap
from .xpcs_result import XpcsResult

logger = logging.getLogger(__name__)


def solve_multitau(*args: Any, **kwargs: Any) -> Union[str, None]:
    kwargs_record = kwargs.copy()
    kwargs_record["analysis_type"] = "multitau"
    return solve_multitau_base(*args, analysis_kwargs=kwargs_record, **kwargs)


def solve_multitau_base(
    qmap: Optional[Union[str, Path]] = None,
    raw: Optional[Union[str, Path]] = None,
    output: str = "cluster_results",
    batch_size: int = 8,
    gpu_id: int = 0,
    verbose: bool = False,
    crop_ratio_threshold: float = 0.5,
    num_loaders: int = 16,
    normalize_frame: bool = False,
    begin_frame: int = 0,
    end_frame: int = -1,
    avg_frame: int = 1,
    stride_frame: int = 1,
    overwrite: bool = False,
    save_G2: bool = False,
    analysis_kwargs: Optional[dict] = None,
    save_results: bool = True,
    num_partial_g2: int = 0,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
    bin_time_s: float = 1e-6,
    run_config_path=None,
    **kwargs: Any,
) -> Union[str, None]:
    log_level = logging.INFO if verbose else logging.ERROR
    logger.setLevel(log_level)

    device = f"cuda:{gpu_id}" if gpu_id >= 0 else "cpu"

    # create qpartitionmap
    qpm = XpcsQPartitionMap(
        qmap, device=device, crop_ratio_threshold=crop_ratio_threshold
    )

    if verbose:
        qpm.describe()
        logger.info(f"device: {device}")

    # create dataset
    dset, use_loader = create_dataset(
        raw,
        device,
        mask_crop=qpm.mask_crop,
        avg_frame=avg_frame,
        begin_frame=begin_frame,
        end_frame=end_frame,
        stride_frame=stride_frame,
        bin_time_s=bin_time_s,
        run_config_path=run_config_path,
    )

    # in some detectors/configurations, the qmap is rotated
    qpm.update_rotation(dset.det_size)

    # determine the metadata path
    # dirname(FILES_IN_CURRENT_FOLDER) gives empty string
    meta_dir = os.path.dirname(os.path.abspath(raw))

    xb = MultitauCorrelator(
        dset.det_size,
        frame_num=dset.frame_num,
        queue_size=batch_size,  # batch_size is the minimal value
        auto_queue=True,
        device=device,
        mask_crop=qpm.mask_crop,
        normalize_frame=normalize_frame,
        qpm=qpm,
        num_partial_g2=num_partial_g2,
    )

    if verbose:
        dset.describe()
        xb.describe()
        logger.info("correlation solver created.")

    t_start = time.perf_counter()
    xb.process_dataset(
        dset, verbose=verbose, use_loader=use_loader, num_workers=num_loaders
    )
    t_end = time.perf_counter()
    t_diff = t_end - t_start
    frequency = dset.frame_num / t_diff
    logger.info(
        f"correlation finished in {t_diff:.2f}s." + f" frequency = {frequency:.2f} Hz"
    )

    t_start = time.perf_counter()
    output_scattering, output_multitau = xb.get_results()
    norm_scattering = qpm.normalize_scattering(output_scattering)
    norm_multitau = qpm.normalize_multitau(output_multitau, save_G2=save_G2)
    part_multitau = xb.get_partial_g2()
    t_end = time.perf_counter()
    logger.info("normalization finished in %.3fs" % (t_end - t_start))

    if save_results:
        with XpcsResult(
            meta_dir,
            qmap,
            output,
            overwrite=overwrite,
            multitau_config=analysis_kwargs,
            rawdata_path=os.path.realpath(raw),
            prefix=prefix,
            suffix=suffix,
        ) as result_file:
            result_file.append(norm_scattering)
            result_file.append(norm_multitau)
            result_file.append(part_multitau)
            if dset.dataset_type == "Timepix4Dataset":
                result_file.correct_t0_for_timepix4(bin_time_s)
        logger.info("multitau analysis finished")
        return result_file.fname
    else:
        result_file_kwargs = {
            "meta_dir": meta_dir,
            "qmap_fname": qmap,
            "output_dir": output,
            "overwrite": overwrite,
            "multitau_config": analysis_kwargs,
        }
        return result_file_kwargs, (norm_scattering, norm_multitau)



import math
import numpy as np
from scipy.sparse import csr_matrix as sp_csr_matrix
import hashlib
import torch


def convert_sparse(a):
    output = np.zeros(shape=(3, a.size), dtype=np.uint32)
    # index
    output[0] = ((a >> 16) & (2 ** 21 - 1)).astype(np.uint32)
    # frame
    output[1] = (a >> 40).astype(np.uint32)
    # count
    output[2] = (a & (2 ** 12 - 1)).astype(np.uint8)
    return output


def gen_tau_bin(frame_num, dpl=4, max_level=60):
    """
    generate tau and fold list according to the frame number
    """
    def worker():
        for n in range(dpl):
            yield n + 1, 0
        level = 0
        while True:
            scl = 2 ** level
            for x in range(dpl + 1, dpl * 2 + 1):
                if x * scl >= frame_num // scl * scl:
                    return
                yield x * scl, level
            level += 1
            # limit the levels to max_level;
            if level > max_level:
                return

    tau_bin = []
    for x, y in worker():
        tau_bin.append([x, y])

    # make it a 2-column array
    tau_bin = np.array(tau_bin, dtype=np.int64).T
    return tau_bin


def sort_tau_bin(tau_bin, frame_num):
    """
    sort the tau_bin object, so tau is relative in each level;
    """
    tau_num = tau_bin.shape[1]
    # rescale tau for each level
    tau_max = np.max(tau_bin[0] // (2 ** tau_bin[1]))

    levels = list(np.unique(tau_bin[1]))
    levels_num = len(levels)
    tau_in_level = {}

    offset = 0
    for level in levels:
        scl = 2 ** level
        tau_list = tau_bin[0][tau_bin[1] == level] // scl

        # tau_idx is used to index the result;
        tau_idx = range(offset, offset + len(tau_list))

        # avg_len is used to compute the average G2, IP, IF
        avg_len = frame_num // scl - tau_list

        tau_in_level[level] = list(zip(tau_list, tau_idx, avg_len))
        offset += len(tau_list)

    assert(tau_num == offset)
    return tau_max, levels, tau_in_level


def is_power_two(num):
    # return true for 1, 2, 4, 8, 16 ...
    while num > 2:
        if num % 2 == 1:
            return False
        num //= 2
    return num == 2 or num == 1


def nonzero_crop(img):
    """
    computes the slice in vertical and horizontal direction to crop the nonzero
        regions of the input array img.
    """
    assert isinstance(img, np.ndarray), 'img must be a numpy.ndarray'
    assert img.ndim == 2, 'img has be a two-dimensional numpy.ndarray'
    idx = np.nonzero(img)
    sl_v = slice(np.min(idx[0]), np.max(idx[0]) + 1)
    sl_h = slice(np.min(idx[1]), np.max(idx[1]) + 1)
    return sl_v, sl_h



import os
import time
import sys
from PyQt5 import uic, Qt
from PyQt5.QtWidgets import (QFileDialog, QMainWindow, QApplication, QWidget,
                             QVBoxLayout, QHBoxLayout, QSplitter)
from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from gpu_corr_server import GPUServer, get_system_information
import json
import torch
from pyqtgraph import DataTreeWidget, ImageView
import traceback
import pyqtgraph as pg
import numpy as np

pg.setConfigOption('imageAxisOrder', 'row-major')

home_dir = os.path.join(os.path.expanduser('~'), '.xpcs_boost_server')
if not os.path.isdir(home_dir):
    os.mkdir(home_dir)


class SetupPreviewer(QSplitter):
    def __init__(self):
        super(SetupPreviewer, self).__init__()
        self.initUI()

    def initUI(self):
        hbox12 = QHBoxLayout()
        self.img1 = ImageView()
        self.img2 = ImageView()
        hbox12.addWidget(self.img1)
        hbox12.addWidget(self.img2)

        hbox34 = QHBoxLayout()
        self.img3 = ImageView()
        self.img4 = ImageView()
        hbox34.addWidget(self.img3)
        hbox34.addWidget(self.img4)
        vbox_img = QVBoxLayout()
        vbox_img.addLayout(hbox12)
        vbox_img.addLayout(hbox34)
        img_widget = QWidget()
        img_widget.setLayout(vbox_img)

        # hbox = QSplitter()
        self.data_tree = DataTreeWidget()
        self.data_tree.setMaximumWidth(400)
        self.addWidget(self.data_tree)
        self.addWidget(img_widget)

        self.setGeometry(300, 300, 1200, 800)
        self.setWindowTitle('Setup Previewer')
        self.show()

    def set_data(self, fname, setup):
        img_hdl = (self.img1, self.img2, self.img3, self.img4)
        for hdl, key in zip(img_hdl, ("qMap", "qCMap", "qRMap", "mask")):
            hdl.setImage(setup[key])
            text1 = pg.TextItem(text=f'{key}', color=(0, 0, 255))
            plot_view = hdl.getView()
            plot_view.addItem(text1)
            setup.pop(key, None)
        self.data_tree.setData(setup)
        self.setWindowTitle(f'Setup Previewer: {fname}')


class Ui(QMainWindow):
    def __init__(self):
        super(Ui, self).__init__()
        uic.loadUi('gpuBoostServer.ui', self)
        self.btn_start.clicked.connect(self.start_server)
        self.btn_stop.clicked.connect(self.stop_server)
        self.home_dir = home_dir

        self.is_running = False
        self.server_kwargs = None
        self.server = None
        self.update_time = 0.500  # 500 ms

        self.timer = QTimer()
        self.timer.timeout.connect(self.update_status)

        self.btn_select_result_dir.clicked.connect(self.select_result_dir)
        self.btn_submit.clicked.connect(self.submit_jobs)
        # plot the load
        self.load_window = 256

        xline = np.arange(self.load_window) * self.update_time
        self.load_xline = xline - np.max(xline)
        self.load_plot.setBackground('w')
        self.load_plot.setLabel('left', 'jobs/s')
        self.load_plot.setLabel('bottom', 'Time', units='s')
        self.load_plot.hideAxis('bottom')
        self.load_plot.setMouseEnabled(x=False, y=False)

        self.load_save_setting(mode="load")
        sys_info = get_system_information()
        self.model = QStandardItemModel()
        for n in range(sys_info['num_gpus']):
            name = sys_info['device'][n]['name']
            label = name + f'_{n}'
            item = QStandardItem(label)
            item.setCheckState(True)
            item.setCheckable(True)
            self.model.appendRow(item)

        self.device_list.setModel(self.model)
        self.show()

    def select_setup_file(self, idx=1):
        if self.is_running:
            self.statusbar.showMessage('Server is running now.', 2000)
            return
        assert idx in (1, 2, 3)
        display = f'setup_fname_{idx}'
        f = QFileDialog.getOpenFileName(self, f'select setup file {idx}')[0]
        if os.path.isfile(f):
            self.__dict__[display].setText(f)
        else:
            self.statusbar.showMessage('the setup file is not valid.')
        return

    def select_result_dir(self):
        if self.is_running:
            self.statusbar.showMessage('Server is running now.', 2000)
            return

        d = QFileDialog.getExistingDirectory(self, f'select result directory')
        if os.path.isdir(d):
            self.result_dir.setText(d)
        else:
            self.statusbar.showMessage('the result directory is not valid.')

    def submit_jobs(self):
        if not self.is_running:
            self.statusbar.showMessage('Server is not running. quit', 2000)
            return

        flists = QFileDialog.getOpenFileName(self,
                                             f'select raw files to reduce',
                                             directory='/clhome/MQICHU',
                                             filter="TXT (*.txt)")
        # filter="TIFF (*.tif);;HDF (*.h5 *.hdf)")

        # self.server.submit_jobs(flists[0])
        # self.server.submit_(flists[0])
        print(flists[0])
        self.server.process_all_files(flists[0])

    def update_status(self):
        if self.server is not None:
            flag, message, jobs_info = self.server.get_status(self.load_window)
        else:
            flag, message, jobs_info = False, "stopped", None

        if flag:
            self.lineEdit_8.setStyleSheet("background-color: rgb(0, 255, 0)")
            yline = jobs_info[:, 3] / self.update_time
            self.load_plot.plot(self.load_xline,
                                yline,
                                fillLevel=0,
                                clear=True,
                                brush=(50, 50, 200, 200))
        else:
            self.lineEdit_8.setStyleSheet("background-color: rgb(255, 255, 0)")
        message = str(time.asctime()) + ' | ' + message
        self.lineEdit_8.setText(message)

    def start_server(self):
        if self.is_running:
            self.statusbar.showMessage('Server is running.', 1000)
            return

        self.statusbar.showMessage('Server is starting', 8000)

        kwargs = {}
        kwargs['port'] = self.port.value()
        kwargs['output'] = self.result_dir.text()
        kwargs['worker_per_gpu'] = 1
        self.server_kwargs = kwargs

        gpu_selection = []
        for n in range(self.model.rowCount()):
            idx = self.model.index(n, 0)
            flag = self.model.itemFromIndex(idx).checkState()
            gpu_selection.append(flag)

        if sum(gpu_selection) == 0:
            self.statusbar.showMessage('None of the GPUs is selected', 1000)
            return
        else:
            kwargs['gpu_selection'] = gpu_selection

        try:
            self.server = GPUServer(**kwargs)
        except Exception as e:
            print('failed to start server')
            traceback.print_exc()

        time.sleep(1)
        self.timer.start(int(self.update_time * 1000))
        self.is_running = True
        self.lineEdit.setText(self.server.ip_addr)
        self.statusbar.showMessage('Server started', 1000)

    def enable_disable_panel(self, flag=True, idx=0):
        pass
        # self.gb_det_1.setEnabled(flag)

    def stop_server(self):
        self.statusbar.showMessage('Server is stopping', 8000)
        if self.is_running:
            self.timer.stop()
            self.server.stop_server()
            self.server = None
            self.is_running = False
        self.update_status()
        self.statusbar.showMessage('Server stopped', 1000)

    def load_save_setting(self, mode='save'):
        text_field = ["watch_pv", "watch_folder", "result_dir"]

        num_field = ["port"]
        assert mode in ("save", "load"), "mode not supported."

        fname = os.path.join(self.home_dir, 'last_setting.json')
        if mode == "save":
            kwargs = {}
            for key in text_field:
                kwargs[key] = self.__dict__[key].text()
            for key in num_field:
                kwargs[key] = self.__dict__[key].value()
            with open(fname, 'w') as f:
                json.dump(kwargs, f, indent=4)

        elif mode == "load":
            try:
                with open(fname, 'r') as f:
                    kwargs = json.load(f)
                for key in text_field:
                    self.__dict__[key].setText(kwargs[key])
                for key in num_field:
                    self.__dict__[key].setValue(kwargs[key])
            except Exception:
                pass

    def closeEvent(self, event):
        self.stop_server()
        self.load_save_setting(mode="save")

        super(QMainWindow, self).closeEvent(event)


if __name__ == '__main__':
    torch.multiprocessing.set_start_method('spawn')
    app = QApplication(sys.argv)
    window = Ui()
    app.exec_()



import struct
import numpy as np
import time
import logging
import os
from tqdm import trange
import torch
from .xpcs_dataset import XpcsDataset

logger = logging.getLogger(__name__)


def read_imm_header(file):
    imm_headformat = "ii32s16si16siiiiiiiiiiiiiddiiIiiI40sf40sf40sf40s" + \
                     "f40sf40sf40sf40sf40sf40sfffiiifc295s84s12s"
    imm_fieldnames = [
        'mode', 'compression', 'date', 'prefix', 'number', 'suffix', 'monitor',
        'shutter', 'row_beg', 'row_end', 'col_beg', 'col_end', 'row_bin',
        'col_bin', 'rows', 'cols', 'bytes', 'kinetics', 'kinwinsize',
        'elapsed', 'preset', 'topup', 'inject', 'dlen', 'roi_number',
        'buffer_number', 'systick', 'pv1', 'pv1VAL', 'pv2', 'pv2VAL', 'pv3',
        'pv3VAL', 'pv4', 'pv4VAL', 'pv5', 'pv5VAL', 'pv6', 'pv6VAL', 'pv7',
        'pv7VAL', 'pv8', 'pv8VAL', 'pv9', 'pv9VAL', 'pv10', 'pv10VAL',
        'imageserver', 'CPUspeed', 'immversion', 'corecotick', 'cameratype',
        'threshhold', 'byte632', 'empty_space', 'ZZZZ', 'FFFF'
    ]

    bindata = file.read(1024)

    imm_headerdat = struct.unpack(imm_headformat, bindata)
    imm_header = dict(zip(imm_fieldnames, imm_headerdat))

    return imm_header


class ImmDataset(XpcsDataset):
    """
    Parameters
    ----------
    filename: string
        path to .imm file
    frames_per_point: integer
        number of frames to return as one datum
    """
    def __init__(
        self,
        *args,
        dtype=np.int16,
        frames_per_point=1,
        **kwargs,
    ):
        super(ImmDataset, self).__init__(*args, dtype=dtype, **kwargs)
        self.dataset_type = "IMM Legacy"
        self.frames_per_point = frames_per_point

        self.toc, self.det_size = self.read_toc()

        # update frame_num and batch info
        self.update_batch_info(self.toc.shape[0])
        self.update_det_size(self.det_size)
        self.fh = None

    def read_toc(self):
        """
        read table of content for IMM datasets
        """
        with open(self.fname, "rb") as f:
            header = read_imm_header(f)
            det_size = (header['rows'], header['cols'])
            self.is_sparse = bool(header['compression'] == 6)
            f.seek(0)
            toc = []  # (start byte, element count) pairs
            while True:
                try:
                    header = read_imm_header(f)
                    cur = f.tell()
                    payload_size = header['dlen'] * \
                        (6 if self.is_sparse else 2)
                    toc.append((cur, header['dlen']))

                    file_pos = payload_size + cur
                    f.seek(file_pos)

                    # Check for end of file.
                    if not f.peek(4):
                        break

                except Exception as err:
                    raise IOError("IMM file is corrupted.")

            return np.array(toc), det_size

    def __reset__(self):
        self.fh = None

    def __getbatch__(self, index):
        if self.fh is None:
            self.fh = open(self.fname, 'rb')

        if self.is_sparse:
            x = self.__get_frame_sparse__(index)
        else:
            x = self.__get_frame_dense__(index)

        if self.mask_crop is not None:
            x = x[:, self.mask_crop]
        return x

    def __get_frame_dense__(self, batch_idx):
        beg, end, size = self.get_raw_index(batch_idx)
        idx_list = np.arange(beg, end, self.stride)
        toc = self.toc[idx_list]
        imgs = []
        for (start_byte, event_num) in toc:
            self.fh.seek(start_byte)
            imgs.append(np.fromfile(self.fh, dtype=np.uint16, count=event_num))
        imgs = np.array(imgs).astype(np.int16)
        x = torch.from_numpy(imgs).to(self.device, non_blocking=True)
        return x

    def __get_frame_sparse__(self, batch_idx):
        beg, end, size = self.get_raw_index(batch_idx)
        idx_list = np.arange(beg, end, self.stride)
        toc = self.toc[idx_list]

        frame = []
        index = []
        count = []
        for n, (start_byte, event_num) in enumerate(toc):
            self.fh.seek(start_byte)
            frame.append(np.zeros(event_num, dtype=np.int16) + n)
            index.append(np.fromfile(self.fh, dtype=np.int32, count=event_num))
            count.append(
                np.fromfile(self.fh, dtype=self.dtype, count=event_num))

        # frame = torch.tensor(np.concatenate(frame), device=self.device)
        frame = torch.from_numpy(np.concatenate(frame)).to(self.device,
                non_blocking=True)
        index = torch.from_numpy(np.concatenate(index)).to(self.device,
                non_blocking=True)
        count = torch.from_numpy(np.concatenate(count)).to(self.device,
                non_blocking=True)

        return self.sparse_to_dense(index, frame, count, size)

    def __del__(self):
        try:
            self.fh.close()
        except Exception:
            pass


def read_data(file_name, batch_size=1024):
    logger.info('start, fname is: %s', os.path.basename(file_name))
    logger.info('dirname is: %s', os.path.dirname(file_name))
    imm = ImmDataset(file_name, batch_size=1024)
    # print(imm.toc[0])
    # print(imm.toc[1])
    # print(len(imm))
    # print(imm[1])
    # print(imm.is_sparse)
    # print(imm.det_size)
    logger.info('toc is generated')
    logger.info('toc length is %d', len(imm.toc))
    tot = 0
    stime = time.perf_counter()
    for n in range(len(imm)):
        x = imm[n]
        # tot += torch.sum(x)
    etime = time.perf_counter()
    logger.info('sum of array is %d', tot)
    t_diff = (etime - stime)
    freq = imm.frame_num / t_diff
    print(f"data traversed: {t_diff:02f}s / {freq:04f}Hz")
    return freq


def test01():
    file_name = "/scratch/xpcs_data_raw/" + \
                "A005_Dragonite_25p_Quiescent_att0_Lq0_001/" + \
                "A005_Dragonite_25p_Quiescent_att0_Lq0_001_00001-20000.imm"
    read_data(file_name)


def test02():
    fname = "/home/8ididata/2021-3/hallinan202111/W2397_S15-SEO-C-S_Lq1_080C_att06_001/W2397_S15-SEO-C-S_Lq1_080C_att06_001_00001-01000.imm"
    read_data(fname)


def test03():
    file_name = "/scratch/xpcs_data_raw/" + \
                "A005_Dragonite_25p_Quiescent_att0_Lq0_001/" + \
                "A005_Dragonite_25p_Quiescent_att0_Lq0_001_00001-20000.imm"
    for batch_size in (2 ** np.arange(12)):
        print(batch_size, read_data(file_name, batch_size))


if __name__ == '__main__':
    test03()



import numpy as np
import logging
import torch
from .xpcs_dataset import XpcsDataset
from .help_functions import convert_sparse
import os


logger = logging.getLogger(__name__)


class RigakuDataset(XpcsDataset):
    """
    Parameters
    ----------
    filename: string
        path to .imm file
    """

    def __init__(self, *args, dtype=np.uint8, total_frames=None, **kwargs):
        super(RigakuDataset, self).__init__(*args, dtype=dtype, **kwargs)
        self.dataset_type = "Rigaku 64bit Binary"
        self.is_sparse = True
        self.dtype = np.uint8
        self.ifc, self.mem_addr = self.read_data(total_frames)
        self.ifc = self.to_device()

    def to_device(self, max_fsize_gb_in_gpu=1.0):
        fsize = os.path.getsize(self.fname) / 1024**3
        if fsize <= max_fsize_gb_in_gpu and self.device.startswith("cuda"):
            device = self.device
        else:
            device = "cpu"
            logger.info(f"{self.fname} is {fsize:.2f} GB, offloading to CPU RAM")

        index = torch.tensor(self.ifc[0].astype(np.int32), device=device)
        frame = torch.tensor(self.ifc[1].astype(np.int32), device=device)
        count = torch.tensor(self.ifc[2].astype(np.uint8), device=device)
        return (index, frame, count)

    def read_data(self, total_frames=None):
        with open(self.fname, "r") as f:
            a = np.fromfile(f, dtype=np.uint64)
            d = convert_sparse(a)
            index, frame, count = d[0], d[1], d[2]
            if total_frames is None:
                total_frames = frame[-1] + 1
            else:
                total_frames = max(total_frames, frame[-1] + 1)

        # fix Rigaku3m module index offset
        if np.min(index) >= 1024 * 1024:
            index -= 1024 * 1024

        # update frame_num and batch_num
        self.update_batch_info(total_frames)

        all_index = np.arange(0, self.frame_num_raw + 2)
        all_index[-1] = self.frame_num_raw
        # build an index map for all indexes
        beg = np.searchsorted(frame, all_index[:-1])
        end = np.searchsorted(frame, all_index[1:])
        mem_addr = [slice(a, b) for a, b in zip(beg, end)]
        return (index, frame, count), mem_addr

    def __getbatch__(self, idx):
        # frame begin and end
        beg, end, size = self.get_raw_index(idx)
        device = self.ifc[0].device
        x = torch.zeros(size, self.pixel_num, dtype=torch.uint8, device=device)
        if self.stride == 1:
            # the data is continuous in RAM; convert by batch
            sla, slb = self.mem_addr[beg], self.mem_addr[end]
            a, b = sla.start, slb.start
            x[self.ifc[1][a:b].long() - beg, self.ifc[0][a:b].long()] = self.ifc[2][a:b]
        else:
            # the data is discrete in RAM; convert frame by frame
            for n, idx in enumerate(np.arange(beg, end, self.stride)):
                sl = self.mem_addr[idx]
                x[n, self.ifc[0][sl].long()] = self.ifc[2][sl]
        x = x.to(self.device)
        if self.mask_crop is not None:
            x = x[:, self.mask_crop]
        return x


def test():
    fname = (
        "/home/beams/MQICHU/ramdisk/O018_Silica_D100_att0_Rq0_00001/"
        "O018_Silica_D100_att0_Rq0_00001.bin"
    )
    ds = RigakuDataset(fname)
    # for n in range(len(ds)):
    #     print(n, ds[n].shape)


if __name__ == "__main__":
    test()



import os
import sys
import json
import torch
import psutil
from PyQt5 import uic, QtCore
from PyQt5.QtWidgets import (QApplication, QMainWindow, QFileDialog, QListView,
                             QAbstractItemView, QTreeView)
from gpu_corr_solver import GPUSolverWorker, get_raw_meta


def get_system_information():
    sys_info = {}
    sys_info['num_gpus'] = torch.cuda.device_count()
    sys_info['num_cpus'] = psutil.cpu_count(logical=False)
    scale = 1024**3

    device = {}
    for n in range(sys_info['num_gpus']):
        a = torch.cuda.get_device_properties('cuda:%d' % n)
        device[n] = {'name': a.name, 'total_memory': a.total_memory / scale}

    cpu_ram = psutil.virtual_memory().total / scale
    # set an upper limit to cpu_ram;
    cpu_ram = min(36.0, cpu_ram)
    device[-1] = {'name': 'cpu', 'total_memory': cpu_ram}
    sys_info['device'] = device

    return sys_info


class TableDataModel(QtCore.QAbstractTableModel):
    def __init__(self, input_list=None, max_display=16384) -> None:
        super().__init__()
        if input_list is None:
            self.input_list = []
        else:
            self.input_list = input_list
        self.max_display = max_display
        self.xlabels = [
            'id', 'priority', 'gpu_id', 'status', 'frames', 'progress(%)',
            'time/freq', 'fname'
        ]

    # overwrite parent method
    def data(self, index, role):
        if role == QtCore.Qt.DisplayRole:
            x = self.input_list[len(self.input_list) - 1 - index.row()]
            ret = x.get_data()
            return ret[index.column()]

    # overwrite parent method
    def rowCount(self, index):
        return min(self.max_display, len(self.input_list))

    # overwrite parent method
    def columnCount(self, index):
        return len(self.xlabels)

    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self.xlabels[section]

    def extend(self, new_input_list):
        self.input_list.extend(new_input_list)
        self.layoutChanged.emit()

    def append(self, new_item):
        self.input_list.append(new_item)
        self.layoutChanged.emit()

    def replace(self, new_input_list):
        self.input_list.clear()
        self.extend(new_input_list)

    def pop(self, index):
        if 0 <= index < self.__len__():
            self.input_list.pop(index)
            self.layoutChanged.emit()

    def __len__(self):
        return len(self.input_list)

    def __getitem__(self, i):
        return self.input_list[i]

    def copy(self):
        return self.input_list.copy()

    def remove(self, x):
        self.input_list.remove(x)

    def clear(self):
        self.input_list.clear()

    def emit(self):
        self.layoutChanged.emit()


class Ui(QMainWindow):
    def __init__(self, config_fname=None):
        super(Ui, self).__init__()
        uic.loadUi('gpu_corr.ui', self)
        self.sys_info = get_system_information()
        self.gpu_id.setMaximum(self.sys_info['num_gpus'] - 1)
        self.num_worker.setMaximum(self.sys_info['num_cpus'])
        self.num_worker.setValue(self.sys_info['num_cpus'] // 2)

        self.config_fname = config_fname
        self.config = None

        # only allow one worker because the computing is intense; jobs will be
        # queued
        self.thread_pool = QtCore.QThreadPool()
        self.thread_pool.setMaxThreadCount(1)
        self.workers = TableDataModel()
        self.job_table.setModel(self.workers)

        self.job_count = 0
        self.last_raw_dir = None
        self.load_last_config()
        self.show()
        self.adjust_gpu_id()

    def load_qmap(self):
        work_dir = os.path.dirname(self.qmap.text())
        if not os.path.isdir(work_dir):
            work_dir = None
        f = QFileDialog.getOpenFileName(self, 'select qmap file',
                directory=work_dir)[0]
        if f in [None, '']:
            return
        self.config['qmap'] = f
        self.qmap.setText(f)

    def load_raw(self):
        # https://stackoverflow.com/questions/38252419
        file_dialog = QFileDialog(self, directory=self.last_raw_dir)
        file_dialog.setFileMode(QFileDialog.DirectoryOnly)
        file_dialog.setOption(QFileDialog.DontUseNativeDialog, True)
        file_view = file_dialog.findChild(QListView, 'listView')

        # to make it possible to select multiple directories:
        if file_view:
            file_view.setSelectionMode(QAbstractItemView.MultiSelection)
        f_tree_view = file_dialog.findChild(QTreeView)
        if f_tree_view:
            f_tree_view.setSelectionMode(QAbstractItemView.MultiSelection)

        paths = None
        if file_dialog.exec():
            paths = file_dialog.selectedFiles()

        if paths in [None, '', []]:
            return

        valid_path = []
        for x in paths:
            raw, meta = get_raw_meta(x)
            if raw is not None and meta is not None:
                valid_path.append(x)

        if len(valid_path) > 0:
            self.config['raw_folder'].extend(valid_path)
            # self.raw_folder.clear()
            self.raw_folder.addItems(valid_path)
            self.last_raw_dir = os.path.dirname(valid_path[-1])

    def set_output(self):
        work_dir = os.path.dirname(self.output.text())
        if not os.path.isdir(work_dir):
            work_dir = None

        f = QFileDialog.getExistingDirectory(self, 'select saving folder',
                directory=work_dir)
        if f in [None, '']:
            return
        self.config['output'] = f
        self.output.setText(f)

    def adjust_gpu_id(self):
        gpu_id = self.gpu_id.value()
        device = self.sys_info['device'][gpu_id]
        self.gpu_id_label.setText('Processing Unit ID (%s)' % device['name'])
        self.max_memory.setMaximum(device['total_memory'] * 0.8)
        self.max_memory.setValue(device['total_memory'] * 0.5)
        return

    def update_table(self):
        self.workers.emit()

    def clear_raw_folders(self):
        self.raw_folder.clear()
        self.config['raw_folder'] = []

    def submit_job(self):
        kwargs = {}
        kwargs.update(self.config)
        raw_folders = kwargs.pop('raw_folder')
        self.raw_folder.clear()
        self.config['raw_folder'] = []

        kwargs['gpu_id'] = self.gpu_id.value()
        kwargs['batch_size'] = int(self.batch_size.currentText())
        kwargs['crop_ratio_threshold'] = self.masked_ratio_threshold.value()
        kwargs['priority'] = self.job_priority.value()
        kwargs['num_worker'] = self.num_worker.value()
        kwargs['max_memory'] = self.max_memory.value()
        if not os.path.isdir(kwargs['output']) or \
           not os.path.isfile(kwargs['qmap']):
            return

        for n in range(len(raw_folders)):
            kw = kwargs.copy()
            kw['raw_folder'] = raw_folders[n]
            self.job_count += 1
            # self.job_producer.submit_job(kwargs)
            worker = GPUSolverWorker(jid=self.job_count, **kw)
            self.workers.append(worker)
            worker.signals.status.connect(self.update_table)
            worker.signals.progress.connect(self.update_table)
            # the thread count is one; so if previous workers are running,
            # the new # woker is queued until the previous one is done;
            self.thread_pool.start(worker, priority=kw['priority'])

    def save_config(self):
        with open(self.config_fname, 'w') as fhdl:
            json.dump(self.config, fhdl)

    def load_last_config(self, ):
        if os.path.isfile(self.config_fname):
            with open(self.config_fname, 'r') as fhdl:
                self.config = json.load(fhdl)
                for key, val in self.config.items():
                    if val is None or key not in [
                            'qmap', 'output', 'raw_folder']:
                        continue
                    if key != 'raw_folder':
                        self.__dict__[key].setText(self.config[key])
                    else:
                        self.__dict__[key].addItems(self.config[key])
                        if len(self.config[key]) > 0:
                            self.last_raw_dir = os.path.dirname(self.config[key][0])
        else:
            self.config = {'qmap': None, 'raw_folder': [], 'output': None}
        return

    def closeEvent(self, e) -> None:
        self.save_config()

    def __del__(self):
        pass


home_dir = os.path.join(os.path.expanduser('~'), '.xpcs_boost')
if not os.path.isdir(home_dir):
    os.mkdir(home_dir)
config_fname = os.path.join(home_dir, 'last_config.json')
app = QApplication([])
window = Ui(config_fname)
app.exec()
